from . import cnn_dann, latent_align, tbigru, mmd, adarul, consistency, dann

from .conditional import ConditionalMmdApproach, ConditionalDannApproach
from .pseudo_labels import generate_pseudo_labels, patch_pseudo_labels
from .latent_align import LatentAlignApproach, LatentAlignFttpApproach
from .tbigru import select_features, VibrationFeatureExtractor
from .mmd import MmdApproach
from .adarul import AdaRulApproach
from .supervised import SupervisedApproach
from .consistency import ConsistencyApproach
from .dann import DannApproach



"""A module of feature extractors based on recurrent neural networks."""

from typing import List, Optional, Tuple, Type, Union

import torch
from torch import nn

from rul_adapt.utils import pairwise


class LstmExtractor(nn.Module):
    """A Long Short Term Memory (LSTM) based network that extracts a feature vector
    from same-length time windows.

    This feature extractor consists of a multi-layer LSTM and an optional fully
    connected (FC) layer with a ReLU activation function. The LSTM layers can be
    configured as bidirectional. Dropout can be applied separately to LSTM and FC
    layers.

    The data flow is as follows: `Input --> LSTM x n --> [FC] --> Output`

    The expected input shape is `[batch_size, num_features, window_size]`.

    Examples:

        Without FC
        >>> import torch
        >>> from rul_adapt.model import LstmExtractor
        >>> lstm = LstmExtractor(input_channels=14,units=[16, 16])
        >>> lstm(torch.randn(10, 14, 30)).shape
        torch.Size([10, 16])

        With FC
        >>> from rul_adapt.model import LstmExtractor
        >>> lstm = LstmExtractor(input_channels=14,units=[16, 16],fc_units=8)
        >>> lstm(torch.randn(10, 14, 30)).shape
        torch.Size([10, 8])
    """

    def __init__(
        self,
        input_channels: int,
        units: List[int],
        fc_units: Optional[int] = None,
        dropout: float = 0.0,
        fc_dropout: float = 0.0,
        bidirectional: bool = False,
    ) -> None:
        """
        Create a new LSTM-based feature extractor.

        The `units` are the output units for each LSTM layer. If `bidirectional`
        is set to `True`, a BiLSTM is used and the output units are doubled. If
        `fc_units` is set, a fully connected layer is appended. The number of output
        features of this network is either `units[-1]` by default,
        `2 * units[ -1]` if bidirectional is set, or `fc_units` if it is set.

        Dropout can be applied to each LSTM layer by setting `lstm_dropout` to a
        number greater than zero. The same is valid for the fully connected layer and
        `fc_dropout`.

        Args:
            input_channels: The number of input channels.
            units: The list of output units for the LSTM layers.
            fc_units: The number of output units for the fully connected layer.
            dropout: The dropout probability for the LSTM layers.
            fc_dropout: The dropout probability for the fully connected layer.
            bidirectional: Whether to use a BiLSTM.
        """
        super().__init__()

        self.input_channels = input_channels
        self.units = units
        self.dropout = dropout
        self.fc_units = fc_units
        self.fc_dropout = fc_dropout
        self.bidirectional = bidirectional

        self._lstm_layers = _get_rnn_layers(
            nn.LSTM,
            self.units,
            self.input_channels,
            self.dropout,
            self.bidirectional,
        )
        self._fc_layer = self._get_fc_layer()

    def _get_fc_layer(self) -> nn.Module:
        fc_layer = nn.Sequential()
        if self.fc_units is not None:
            if self.fc_dropout > 0:
                fc_layer.append(nn.Dropout(self.fc_dropout))
            fc_in_units = self.units[-1] * (2 if self.bidirectional else 1)
            fc_layer.append(nn.Linear(fc_in_units, self.fc_units))
            fc_layer.append(nn.ReLU())

        return fc_layer

    def forward(self, inputs: torch.Tensor) -> torch.Tensor:
        inputs = torch.permute(inputs, (2, 0, 1))
        inputs, _ = self._lstm_layers(inputs)
        if self.bidirectional:
            forward, backward = torch.split(inputs, self.units[-1], dim=2)
            inputs = torch.cat([forward[-1], backward[0]], dim=1)
        else:
            inputs = inputs[-1]
        inputs = self._fc_layer(inputs)

        return inputs


class GruExtractor(nn.Module):
    """A Gated Recurrent Unit (GRU) based network that extracts a feature vector
    from same-length time windows.

    This feature extractor consists of multiple fully connected (FC) layers with
    a ReLU activation functions and a multi-layer GRU. The GRU layers can be
    configured as bidirectional. Dropout can be applied separately to the GRU
    layers.

    The data flow is as follows: `Input --> FC x n --> GRU x m --> Output`

    The expected input shape is `[batch_size, num_features, window_size]`.

    Examples:
        >>> import torch
        >>> from rul_adapt.model import GruExtractor
        >>> gru = GruExtractor(input_channels=14, fc_units=[16, 8], gru_units=[8])
        >>> gru(torch.randn(10, 14, 30)).shape
        torch.Size([10, 8])
    """

    def __init__(
        self,
        input_channels: int,
        fc_units: List[int],
        gru_units: List[int],
        gru_dropout: float = 0.0,
        bidirectional: bool = False,
    ) -> None:
        """
        Create a new GRU-based feature extractor.

        The `fc_units` are the output units for each fully connected layer
        and `gru_units` for each LSTM layer. If `bidirectional` is set to `True`,
        a BiGRU is used and the output units are doubled. The number of output
        features of this network is either `gru_units[-1]` by default, or `2 *
        gru_units[ -1]` if bidirectional is set.

        Dropout can be applied to each GRU layer by setting `lstm_dropout` to a
        number greater than zero.

        Args:
            input_channels: The number of input channels.
            fc_units: The list of output units for the fully connected layers.
            gru_units: The list of output units for the GRU layers
            gru_dropout: The dropout probability for the GRU layers.
            bidirectional: Whether to use a BiGRU.
        """
        super().__init__()

        self.input_channels = input_channels
        self.gru_units = gru_units
        self.gru_dropout = gru_dropout
        self.fc_units = fc_units
        self.bidirectional = bidirectional

        self._fc_layer = self._get_fc_layers()
        self._gru_layers = _get_rnn_layers(
            nn.GRU,
            self.gru_units,
            self.fc_units[-1],
            self.gru_dropout,
            self.bidirectional,
        )

    def _get_fc_layers(self) -> nn.Module:
        fc_layer = nn.Sequential()
        fc_units = [self.input_channels] + self.fc_units
        for in_units, out_units in pairwise(fc_units):
            fc_layer.append(nn.Conv1d(in_units, out_units, kernel_size=1))
            fc_layer.append(nn.ReLU())

        return fc_layer

    def forward(self, inputs: torch.Tensor) -> torch.Tensor:
        inputs = self._fc_layer(inputs)  # 1x1 conv as linear layer on each time step
        inputs = torch.permute(inputs, (2, 0, 1))
        inputs, _ = self._gru_layers(inputs)
        if self.bidirectional:
            forward, backward = torch.split(inputs, self.gru_units[-1], dim=2)
            inputs = torch.cat([forward[-1], backward[0]], dim=1)
        else:
            inputs = inputs[-1]

        return inputs


def _get_rnn_layers(
    rnn_func: Union[Type[nn.LSTM], Type[nn.GRU]],
    num_units: List[int],
    input_channels: int,
    dropout: float,
    bidirectional: bool,
) -> nn.Module:
    layers: nn.Module
    if all(num_units[0] == u for u in num_units):
        layers = rnn_func(
            input_channels,
            num_units[0],
            num_layers=len(num_units),
            dropout=dropout,
            bidirectional=bidirectional,
        )
    else:
        layers = _Rnn(rnn_func, input_channels, num_units, dropout, bidirectional)

    return layers


class _Rnn(nn.Module):
    def __init__(
        self,
        rnn_func: Union[Type[nn.LSTM], Type[nn.GRU]],
        input_channels: int,
        rnn_units: List[int],
        dropout: float,
        bidirectional: bool,
    ) -> None:
        super().__init__()

        self._rnn_func = rnn_func
        self.input_channels = input_channels
        self.rnn_units = rnn_units
        self.dropout = dropout
        self.bidirectional = bidirectional

        if not self.rnn_units:
            raise ValueError(f"Cannot build stacked {self._rnn_func} with zero layers.")
        self._layers = self._get_layers()

    def _get_layers(self) -> nn.ModuleList:
        lstm_layers = nn.ModuleList()  # registers items as parameters
        #
        unit_multiplier = 2 if self.bidirectional else 1  # 2x in channels for BiRnn
        unit_iter = pairwise([self.input_channels / unit_multiplier] + self.rnn_units)
        for input_channels, num_units in unit_iter:
            lstm_layers.append(
                self._rnn_func(
                    int(input_channels * unit_multiplier),  # in channels could be odd
                    num_units,
                    bidirectional=self.bidirectional,
                )
            )

        return lstm_layers

    def forward(
        self, inputs: torch.Tensor
    ) -> Tuple[torch.Tensor, Tuple[torch.Tensor, torch.Tensor]]:
        for i, rnn in enumerate(self._layers, start=1):
            inputs, states = rnn(inputs)
            if i < len(self._layers):  # no dropout on last layer outputs
                inputs = nn.functional.dropout(inputs, self.dropout)

        return inputs, states



from typing import Callable, List, Literal, Tuple

import torch
import torchmetrics
from torch import nn

import rul_adapt


class AdaptionEvaluator(nn.Module):
    def __init__(
        self,
        network_func: Callable[[torch.Tensor], torch.Tensor],
        log_func: Callable[[str, torchmetrics.Metric], None],
        score_mode: Literal["phm08", "phm12"] = "phm08",
        degraded_only: bool = False,
    ):
        super().__init__()

        self.network_func = network_func
        self.log_func = log_func
        self.score_mode = score_mode
        self.degraded_only = degraded_only

        self.val_metrics = self._get_default_metrics()
        self.test_metrics = self._get_default_metrics()

    def _get_default_metrics(self) -> nn.ModuleDict:
        return nn.ModuleDict(
            {
                "source": nn.ModuleDict(
                    {
                        "rmse": torchmetrics.MeanSquaredError(squared=False),
                        "score": rul_adapt.loss.RULScore(self.score_mode),
                    }
                ),
                "target": nn.ModuleDict(
                    {
                        "rmse": torchmetrics.MeanSquaredError(squared=False),
                        "score": rul_adapt.loss.RULScore(self.score_mode),
                    }
                ),
            }
        )

    def validation(
        self, batch: List[torch.Tensor], domain: Literal["source", "target"]
    ) -> None:
        self._evaluate("val", self.val_metrics, batch, domain)

    def test(
        self, batch: List[torch.Tensor], domain: Literal["source", "target"]
    ) -> None:
        self._evaluate("test", self.test_metrics, batch, domain)

    def _evaluate(
        self,
        prefix: str,
        metrics: nn.ModuleDict,
        batch: List[torch.Tensor],
        domain: Literal["source", "target"],
    ) -> None:
        self._check_domain(domain, prefix)
        features, labels = batch
        features, labels = filter_batch(features, labels, self.degraded_only)
        labels = labels[:, None]
        predictions = self.network_func(features)
        for metric_name, metric in metrics[domain].items():
            metric(predictions, labels)
            self.log_func(f"{prefix}/{domain}/{metric_name}", metric)

    def _check_domain(self, domain: str, prefix: str) -> None:
        if domain not in ["source", "target"]:
            raise RuntimeError(
                f"Unexpected {prefix} domain '{domain}'. "
                "Use either 'source' or 'target'."
            )


def filter_batch(
    features: torch.Tensor, labels: torch.Tensor, degraded_only: bool = False
) -> Tuple[torch.Tensor, torch.Tensor]:
    if degraded_only:
        if torch.any(labels > 1.0):
            raise RuntimeError(
                "Degradation-only evaluation configured which works only with "
                "normalized RUL, but labels contain values greater than 1.0."
            )
        degraded = labels < 1.0
        features = features[degraded]
        labels = labels[degraded]

    return features, labels



from .adaption import (
    MaximumMeanDiscrepancyLoss,
    JointMaximumMeanDiscrepancyLoss,
    DomainAdversarialLoss,
    ConsistencyLoss,
)
from .alignment import (
    HealthyStateAlignmentLoss,
    DegradationDirectionAlignmentLoss,
    DegradationLevelRegularizationLoss,
)
from .rul import rul_score, RULScore
from .conditional import ConditionalAdaptionLoss



from typing import Tuple, Any

import hydra
import omegaconf
import pytorch_lightning as pl
import rul_datasets

from rul_adapt.approach import MmdApproach


def get_tbigru(
    source_fd: int, target_fd: int, **trainer_kwargs: Any
) -> Tuple[rul_datasets.DomainAdaptionDataModule, MmdApproach, pl.Trainer]:
    """
    Construct a TBiGRU approach for FEMTO with the original hyperparameters.

    Examples:
        ```pycon
        >>> import rul_adapt
        >>> dm, tbigru, trainer = rul_adapt.construct.get_tbigru(3, 1)
        >>> trainer.fit(tbigru, dm)
        >>> trainer.test(tbigru, dm)
        ```

    Args:
        source_fd: The source FD of FEMTO.
        target_fd: The target FD of FEMTO.
        trainer_kwargs: Overrides for the trainer class.
    Returns:
        dm: The data module for adaption of two FEMTO sub-datasets.
        dann: The TBiGRU approach with feature extractor and regressor.
        trainer: The trainer object.
    """
    config = get_tbigru_config(source_fd, target_fd)
    dm, dann, trainer = tbigru_from_config(config, **trainer_kwargs)

    return dm, dann, trainer


def get_tbigru_config(source_fd: int, target_fd: int) -> omegaconf.DictConfig:
    """
    Get a configuration for the TBiGRU approach.

    The configuration can be modified and fed to [tbigru_from_config]
    [rul_adapt.construct.tbigru.tbigru_from_config] to create the approach.

    Args:
        source_fd: The source FD of FEMTO.
        target_fd: The target FD of FEMTO.
    Returns:
        The TBiGRU configuration.
    """
    _validate(source_fd, target_fd)
    with hydra.initialize("config", version_base="1.1"):
        config = hydra.compose("base", overrides=[f"+task={source_fd}-{target_fd}"])

    return config


def tbigru_from_config(
    config: omegaconf.DictConfig, **trainer_kwargs: Any
) -> Tuple[rul_datasets.DomainAdaptionDataModule, MmdApproach, pl.Trainer]:
    """
    Construct a TBiGRU approach from a configuration.
    The configuration can be created by calling [get_tbigru_config]
    [rul_adapt.construct.tbigru.get_tbigru_config].

    Args:
        config: The TBiGRU configuration.
        trainer_kwargs: Overrides for the trainer class.
    Returns:
        dm: The data module for adaption of two FEMTO sub-datasets.
        dann: The TBiGRU approach with feature extractor and regressor.
        trainer: The trainer object.
    """
    source = hydra.utils.instantiate(config.dm.source)
    target = hydra.utils.instantiate(config.dm.target)

    source.prepare_data()  # needed in case this is the first use of FEMTO
    target.prepare_data()
    extractor = hydra.utils.instantiate(config.dm.feature_extractor)
    extractor.fit(source.load_split("dev")[0] + target.load_split("dev")[0])

    dm = rul_datasets.DomainAdaptionDataModule(
        rul_datasets.RulDataModule(
            source, config.dm.batch_size, extractor, config.dm.window_size
        ),
        rul_datasets.RulDataModule(
            target, config.dm.batch_size, extractor, config.dm.window_size
        ),
    )

    feature_extractor = hydra.utils.instantiate(config.feature_extractor)
    regressor = hydra.utils.instantiate(config.regressor)

    tbigru = hydra.utils.instantiate(config.tbigru)
    tbigru.set_model(feature_extractor, regressor)

    trainer = hydra.utils.instantiate(config.trainer, **trainer_kwargs)

    return dm, tbigru, trainer


def _validate(source_fd: int, target_fd: int) -> None:
    if source_fd == target_fd:
        raise ValueError(
            f"No configuration for adapting from FD{source_fd:03} to itself."
        )
    elif 1 > source_fd or source_fd > 3:
        raise ValueError(f"FEMTO has only FD001 to FD003 but no FD{source_fd:03}")
    elif 1 > target_fd or target_fd > 3:
        raise ValueError(f"FEMTO has only FD001 to FD003 but no FD{target_fd:03}")



"""Pseudo labeling is a simple approach that takes a model trained on the source
domain to label the target domain. Afterward, the model training is continued on the
combined source and target data. This process is repeated until the validation loss
converges.

Used In:
    * Wang et al. (2022). **Residual Life Prediction of Bearings Based on SENet-TCN
    and Transfer Learning.** *IEEE Access*, *10*,
    [10.1109/ACCESS.2022.3223387](https://doi.org/10.1109/ACCESS.2022.3223387)

Examples:
    ```python
    import torch
    import rul_datasets
    import pytorch_lightning as pl

    from rul_adapt import model
    from rul_adapt import approach

    feat_ex = model.CnnExtractor(14, [16], 30, fc_units=16)
    reg = model.FullyConnectedHead(16, [1])

    supervised = approach.SupervisedApproach(0.001, "rmse", "adam")
    supervised.set_model(feat_ex, reg)

    fd1 = rul_datasets.RulDataModule(rul_datasets.CmapssReader(1), 32)
    fd1.setup()
    fd3 = rul_datasets.RulDataModule(rul_datasets.CmapssReader(3), 32)
    fd3.setup()

    trainer = pl.Trainer(max_epochs=10)
    trainer.fit(supervised, fd3)

    pseudo_labels = approach.generate_pseudo_labels(fd1, supervised)
    pseudo_labels = [max(0, min(125, pl)) for pl in pseudo_labels]
    approach.patch_pseudo_labels(fd3, pseudo_labels)

    combined_data = torch.utils.data.ConcatDataset(
        [fd1.to_dataset("dev"), fd3.to_dataset("dev")]
    )
    combined_dl = torch.utils.data.DataLoader(combined_data, batch_size=32)

    trainer = pl.Trainer(max_epochs=10)
    trainer.fit(supervised, train_dataloader=combined_dl)
    ```
"""

import warnings
from typing import List, Tuple, Optional, Union

import numpy as np
import rul_datasets
import torch


def generate_pseudo_labels(
    dm: rul_datasets.RulDataModule, model: torch.nn.Module, inductive: bool = False
) -> List[float]:
    """
    Generate pseudo labels for the dev set of a data module.

    The pseudo labels are generated for the last timestep of each run. They are
    returned raw and may therefore contain values bigger than `max_rul` or negative
    values. It is recommended to clip them to zero and `max_rul` respectively before
    using them to patch a reader.

    Args:
        dm: The data module to generate pseudo labels for.
        model: The model to use for generating the pseudo labels.
        inductive: Whether to generate pseudo labels for inductive adaption,
                   i.e., use 'test' instead of 'dev' split.

    Returns:
        A list of pseudo labels for the dev set of the data module.
    """
    features, _ = dm.load_split("test" if inductive else "dev", alias="dev")
    last_timestep = torch.stack([f[-1] for f in features])
    pseudo_labels = model(last_timestep).squeeze(axis=1)

    max_rul = get_max_rul(dm.reader)
    if torch.any(pseudo_labels > max_rul):
        warnings.warn(
            "At least one of the generated pseudo labels is greater "
            "than the maximum RUL of the dataset. This may lead to unexpected results "
            "when trying to instantiate a reader with these pseudo labels."
        )
    if torch.any(pseudo_labels < 0):
        warnings.warn(
            "At least one of the generated pseudo labels is negative. "
            "Please consider clipping them to zero."
        )

    return pseudo_labels.tolist()


def patch_pseudo_labels(
    dm: rul_datasets.RulDataModule, pseudo_labels: List[float], inductive: bool = False
) -> None:
    """
    Patch a data module with pseudo labels in-place.

    The pseudo labels are used to replace the RUL targets of the dev set of the data
    module. The validation and test sets are not affected.

    It is not possible to patch the same data module multiple times. Instead,
    instantiate a fresh data module and patch that one.

    Args:
        dm: The data module to patch.
        pseudo_labels: The pseudo labels to use for patching the data module.
        inductive: Whether to generate pseudo labels for inductive adaption,
                   i.e., use 'test' instead of 'dev' split.
    """
    if isinstance(dm.reader, _PseudoLabelReader):
        raise RuntimeError(
            "The data module has already been patched with pseudo labels. "
            "Please instantiate a fresh one."
        )
    dm._reader = _PseudoLabelReader(dm.reader, pseudo_labels, inductive)
    dm.setup()  # overwrites any previously loaded data


class _PseudoLabelReader(rul_datasets.reader.AbstractReader):
    """Reader to wrap another reader and replace the RUL targets of the dev set."""

    def __init__(
        self,
        reader: rul_datasets.reader.AbstractReader,
        pseudo_labels: List[float],
        inductive: bool,
    ) -> None:
        super().__init__(
            reader.fd,
            reader.window_size,
            reader.max_rul,
            reader.percent_broken,
            reader.percent_fail_runs,
            reader.truncate_val,
            reader.truncate_degraded_only,
        )

        if any(pl < 0 for pl in pseudo_labels):
            raise ValueError("Pseudo labels must be non-negative.")
        max_rul = get_max_rul(reader)
        if any(pl > max_rul for pl in pseudo_labels):
            raise ValueError(
                "Pseudo labels must be smaller than the maximum RUL "
                f"of the dataset, {max_rul}."
            )

        self._reader = reader
        self._pseudo_labels = pseudo_labels
        self._inductive = inductive
        self._max_rul = max_rul

    @property
    def dataset_name(self) -> str:
        return self._reader.dataset_name

    @property
    def _patched_split(self) -> str:
        return "test" if self._inductive else "dev"

    @property
    def first_time_to_predict(self) -> Optional[List[int]]:
        if hasattr(self._reader, "first_time_to_predict"):
            return self._reader.first_time_to_predict
        else:
            return None

    @property
    def _dev_fttp(self) -> Optional[List[int]]:
        if self.first_time_to_predict is None:
            return None
        else:
            split = self._patched_split
            return [
                self.first_time_to_predict[i - 1]
                for i in self._reader._preparator.run_split_dist[split]  # type: ignore
            ]

    @property
    def norm_rul(self) -> bool:
        if hasattr(self._reader, "norm_rul"):
            return self._reader.norm_rul
        else:
            return False

    @property
    def fds(self) -> List[int]:
        return self._reader.fds

    def prepare_data(self) -> None:
        self._reader.prepare_data()

    def default_window_size(self, fd: int) -> int:
        return self._reader.default_window_size(fd)

    def load_split(
        self, split: str, alias: Optional[str] = None
    ) -> Tuple[List[np.ndarray], List[np.ndarray]]:
        alias = alias or split
        features, targets = self._reader.load_split(split, alias)
        if split == self._patched_split and alias == "dev":
            targets = self._get_pseudo_labels(features)

        return features, targets

    def load_complete_split(
        self, split: str, alias: str
    ) -> Tuple[List[np.ndarray], List[np.ndarray]]:
        return self._reader.load_complete_split(split, alias)

    def _get_pseudo_labels(self, features: List[np.ndarray]) -> List[np.ndarray]:
        if not len(features) == len(self._pseudo_labels):
            raise RuntimeError(
                f"Number of runs ({len(features)}) and pseudo labels "
                f"({len(self._pseudo_labels)}) do not match."
            )

        if self._dev_fttp is None:
            targets = [
                self._expand_pseudo_label(f, pl)
                for f, pl in zip(features, self._pseudo_labels)
            ]
        else:
            targets = [
                self._expand_pseudo_label_fttp(f, pl, fttp)
                for f, pl, fttp in zip(features, self._pseudo_labels, self._dev_fttp)
            ]

        return targets

    def _expand_pseudo_label(
        self, feature: np.ndarray, pseudo_label: float
    ) -> np.ndarray:
        first_rul = pseudo_label + len(feature)
        rul_values = np.arange(pseudo_label, first_rul)[::-1]
        if self.max_rul is not None:
            rul_values = np.minimum(self.max_rul, rul_values)

        return rul_values

    def _expand_pseudo_label_fttp(
        self, feature: np.ndarray, pseudo_label: float, fttp: int
    ) -> np.ndarray:
        if self.norm_rul:
            fttp -= 1  # linspace includes 1.0 so decrease starts earlier
            rul_values = np.empty(len(feature))
            rul_values[:fttp] = 1.0
            remaining_steps = max(0, len(feature) - fttp)
            rul_values[fttp:] = np.linspace(1.0, pseudo_label, remaining_steps)
        else:
            first_rul = pseudo_label + len(feature)
            max_rul = first_rul - fttp
            rul_values = np.arange(pseudo_label, first_rul)[::-1]
            rul_values = np.minimum(rul_values, max_rul)

        return rul_values


def get_max_rul(reader: rul_datasets.reader.AbstractReader) -> Union[int, float]:
    """Resolve the maximum RUL of a reader to be comparable to floats."""
    return reader.max_rul or (
        1 if hasattr(reader, "norm_rul") and reader.norm_rul else float("inf")
    )



"""The supervised approach trains solely on the labeled source domain. It can be used
for pre-training or as a baseline to compare adaption approaches against.

```python
Data --> FeatureExtractor --> Features --> Regressor  --> RUL Prediction
```
"""

from typing import Literal, Any, List

import torch
import torchmetrics

from rul_adapt import utils
from rul_adapt.approach.abstract import AdaptionApproach
from rul_adapt.approach.evaluation import filter_batch, AdaptionEvaluator


class SupervisedApproach(AdaptionApproach):
    """The supervised approach uses either MSE, MAE or RMSE loss to train a feature
    extractor and regressor in a supervised fashion on the source domain. It can be
    used either for pre-training or as a baseline to compare adaption approaches
    against.

    The regressor needs the same number of input units as the feature extractor has
    output units.

    Examples:
        >>> from rul_adapt import model
        >>> from rul_adapt import approach
        >>> feat_ex = model.CnnExtractor(1, [16, 16, 1], 10, fc_units=16)
        >>> reg = model.FullyConnectedHead(16, [1])
        >>> disc = model.FullyConnectedHead(16, [8, 1], act_func_on_last_layer=False)
        >>> main = approach.SupervisedApproach("mse")
        >>> main.set_model(feat_ex, reg, disc)
    """

    def __init__(
        self,
        loss_type: Literal["mse", "mae", "rmse"],
        rul_scale: int = 1,
        rul_score_mode: Literal["phm08", "phm12"] = "phm08",
        evaluate_degraded_only: bool = False,
        **optim_kwargs: Any,
    ) -> None:
        """
        Create a supervised approach.

        The regressor output can be scaled with `rul_scale` to control its
        magnitude. By default, the RUL values are not scaled.

        For more information about the possible optimizer keyword arguments,
        see [here][rul_adapt.utils.OptimizerFactory].

        Args:
            rul_score_mode:
            loss_type: Training loss function to use. Either 'mse', 'mae' or 'rmse'.
            rul_scale: Scalar to multiply the RUL prediction with.
            evaluate_degraded_only: Whether to only evaluate the RUL score on degraded
                                    samples.
            **optim_kwargs: Keyword arguments for the optimizer, e.g. learning rate.
        """
        super().__init__()

        self.loss_type = loss_type
        self.rul_scale = rul_scale
        self.rul_score_mode = rul_score_mode
        self.evaluate_degraded_only = evaluate_degraded_only
        self.optim_kwargs = optim_kwargs

        self.train_loss = utils.get_loss(loss_type)
        self._get_optimizer = utils.OptimizerFactory(**self.optim_kwargs)
        self.val_loss = torchmetrics.MeanSquaredError(squared=False)
        self.test_loss = torchmetrics.MeanSquaredError(squared=False)
        self.evaluator = AdaptionEvaluator(
            self.forward, self.log, self.rul_score_mode, self.evaluate_degraded_only
        )

        self.save_hyperparameters()

    def configure_optimizers(self) -> utils.OptimizerLRSchedulerConfig:
        return self._get_optimizer(self.parameters())

    def forward(self, inputs: torch.Tensor) -> torch.Tensor:
        return self.regressor(self.feature_extractor(inputs)) * self.rul_scale

    def training_step(self, batch: List[torch.Tensor], batch_idx: int) -> torch.Tensor:
        """
        Execute one training step.

        The `batch` argument is a list of two tensors representing features and
        labels. The features are used to predict RUL values that are compared against
        the labels with the specified training loss. The loss is then logged.

        Args:
            batch: A list of feature and label tensors.
            batch_idx: The index of the current batch.
        Returns:
            The training loss.
        """
        inputs, labels = batch
        predictions = self.forward(inputs)
        loss = self.train_loss(predictions, labels[:, None])
        self.log("train/loss", self.train_loss)

        return loss

    def validation_step(
        self,
        batch: List[torch.Tensor],
        batch_idx: int,
        dataloader_idx: int = -1,
    ) -> None:
        """
        Execute one validation step.

        The `batch` argument is a list of two tensors representing features and
        labels. The features are used to predict RUL values that are compared against
        the labels with an RMSE loss. The loss is then logged.

        Args:
            batch: A list of feature and label tensors.
            batch_idx: The index of the current batch.
        """
        if dataloader_idx == -1:
            self._no_adapt_validation_step(batch)
        else:
            domain = utils.dataloader2domain(dataloader_idx)
            self.evaluator.validation(batch, domain)

    def _no_adapt_validation_step(self, batch: List[torch.Tensor]) -> None:
        inputs, labels = batch
        inputs, labels = filter_batch(
            inputs, labels, degraded_only=self.evaluate_degraded_only
        )
        predictions = self.forward(inputs)
        self.val_loss(predictions, labels[:, None])
        self.log("val/loss", self.val_loss)

    def test_step(
        self, batch: List[torch.Tensor], batch_idx: int, dataloader_idx: int = -1
    ) -> None:
        """
        Execute one test step.

        The `batch` argument is a list of two tensors representing features and
        labels. A RUL prediction is made from the features and the validation RMSE
        and RUL score are calculated. The metrics recorded for dataloader idx zero
        are assumed to be from the source domain and for dataloader idx one from the
        target domain. The metrics are written to the configured logger under the
        prefix `test`.

        Args:
            batch: A list containing a feature and a label tensor.
            batch_idx: The index of the current batch.
            dataloader_idx: The index of the current dataloader (0: source, 1: target).
        """
        if dataloader_idx == -1:
            self._no_adapt_test_step(batch)
        else:
            domain = utils.dataloader2domain(dataloader_idx)
            self.evaluator.test(batch, domain)

    def _no_adapt_test_step(self, batch: List[torch.Tensor]) -> None:
        inputs, labels = batch
        inputs, labels = filter_batch(
            inputs, labels, degraded_only=self.evaluate_degraded_only
        )
        predictions = self.forward(inputs)
        self.test_loss(predictions, labels[:, None])
        self.log("test/loss", self.test_loss)



from typing import Type

import torch
from torch import nn


class ActivationDropoutWrapper(nn.Module):
    def __init__(
        self,
        wrapped: nn.Module,
        act_func: Type[nn.Module] = nn.ReLU,
        dropout: float = 0.5,
    ) -> None:
        super().__init__()

        self.wrapped = wrapped
        self.act_func = act_func
        self.dropout = dropout

        self._act_func = self.act_func()
        self._dropout = nn.Dropout(self.dropout)

    def forward(self, inputs: torch.Tensor) -> torch.Tensor:
        return self._dropout(self._act_func(self.wrapped(inputs)))


class DropoutPrefix(nn.Module):
    def __init__(
        self,
        wrapped: nn.Module,
        dropout: float = 0.5,
    ) -> None:
        super().__init__()

        self.wrapped = wrapped
        self.dropout = dropout

        self._dropout = nn.Dropout(self.dropout)

    def forward(self, inputs: torch.Tensor) -> torch.Tensor:
        return self.wrapped(self._dropout(inputs))



"""A module for network working as a regression or classification head."""

from typing import List, Type, Optional

import torch
from torch import nn

from rul_adapt import utils
from rul_adapt.utils import pairwise


class FullyConnectedHead(nn.Module):
    """A fully connected (FC) network that can be used as a RUL regressor or a domain
    discriminator.

    This network is a stack of fully connected layers with ReLU activation functions
    by default. The activation function can be customized through the `act_func`
    parameter. If the last layer of the network should not have an activation
    function, `act_func_on_last_layer` can be set to `False`.

    The data flow is as follows: `Inputs --> FC x n --> Outputs`

    The expected input shape is `[batch_size, num_features]`.

    Examples:
        Default
        >>> import torch
        >>> from rul_adapt.model import FullyConnectedHead
        >>> regressor = FullyConnectedHead(32, [16, 1])
        >>> outputs = regressor(torch.randn(10, 32))
        >>> outputs.shape
        torch.Size([10, 1])
        >>> type(outputs.grad_fn)
        <class 'ReluBackward0'>

        Custom activation function
        >>> import torch
        >>> from rul_adapt.model import FullyConnectedHead
        >>> regressor = FullyConnectedHead(32, [16, 1], act_func=torch.nn.Sigmoid)
        >>> outputs = regressor(torch.randn(10, 32))
        >>> type(outputs.grad_fn)
        <class 'SigmoidBackward0'>

        Without activation function on last layer
        >>> import torch
        >>> from rul_adapt.model import FullyConnectedHead
        >>> regressor = FullyConnectedHead(32, [16, 1], act_func_on_last_layer=False)
        >>> outputs = regressor(torch.randn(10, 32))
        >>> type(outputs.grad_fn)
        <class 'AddmmBackward0'>
    """

    def __init__(
        self,
        input_channels: int,
        units: List[int],
        dropout: float = 0.0,
        act_func: Type[nn.Module] = nn.ReLU,
        act_func_on_last_layer: bool = True,
    ) -> None:
        """
        Create a new fully connected head network.

        The `units` are the number of output units for each FC layer. The number of
        output features is `units[-1]`. If dropout is used, it is applied in *each*
        layer, including input.

        Args:
            input_channels: The number of input channels.
            units: The number of output units for the FC layers.
            dropout: The dropout probability before each layer. Set to zero to
                     deactivate.
            act_func: The activation function for each layer.
            act_func_on_last_layer: Whether to add the activation function to the last
                                    layer.
        """
        super().__init__()

        self.input_channels = input_channels
        self.units = units
        self.dropout = dropout
        self.act_func: Type[nn.Module] = utils.str2callable(  # type: ignore[assignment]
            act_func, restriction="torch.nn"
        )
        self.act_func_on_last_layer = act_func_on_last_layer

        if not self.units:
            raise ValueError("Cannot build head network with no layers.")

        self._layers = self._get_layers()

    def _get_layers(self) -> nn.Module:
        units = [self.input_channels] + self.units
        act_funcs: List[Optional[Type[nn.Module]]] = [self.act_func] * len(self.units)
        act_funcs[-1] = self.act_func if self.act_func_on_last_layer else None

        layers = nn.Sequential()
        for (in_units, out_units), act_func in zip(pairwise(units), act_funcs):
            layers.append(self._get_fc_layer(in_units, out_units, act_func))

        return layers

    def _get_fc_layer(
        self, in_units: int, out_units: int, act_func: Optional[Type[nn.Module]]
    ) -> nn.Module:
        layer: nn.Module
        if self.dropout > 0:
            layer = nn.Sequential(nn.Dropout(self.dropout))
        else:
            layer = nn.Sequential()

        layer.append(nn.Linear(in_units, out_units))
        if act_func is not None:
            layer.append(act_func())

        return layer

    def forward(self, inputs: torch.Tensor) -> torch.Tensor:
        return self._layers(inputs)



"""A module for conditional unsupervised domain adaption losses."""

from typing import List, Tuple, Sequence

import torch
import torchmetrics
from torch import nn


class ConditionalAdaptionLoss(torchmetrics.Metric):
    """The Conditional Adaptions loss is a combination of multiple losses, each of
    which is only applied to a subset of the incoming data.

    The subsets are defined by fuzzy sets with a rectangular membership function. The
    prediction for each sample is checked against the fuzzy sets, and the
    corresponding loss is applied to the sample. The combined loss can be set as the
    sum of all components or their mean."""

    is_differentiable = True
    higher_is_better = False
    full_state_update = True

    loss: torch.Tensor
    batch_counter: torch.Tensor

    def __init__(
        self,
        adaption_losses: Sequence[torchmetrics.Metric],
        fuzzy_sets: List[Tuple[float, float]],
        mean_over_sets: bool = True,
    ) -> None:
        """
        Create a new Conditional Adaption loss over fuzzy sets.

        The fuzzy sets' boundaries are inclusive on the left and exclusive on the right.
        The left boundary is supposed to be smaller than the right boundary.

        This loss should not be used as a way to accumulate its value over multiple
        batches.

        Args:
            adaption_losses: The list of losses to be applied to the subsets.
            fuzzy_sets: The fuzzy sets to be used for subset membership.
            mean_over_sets: Whether to take the mean or the sum of the losses.
        """
        super().__init__()

        self._check_fuzzy_sets(fuzzy_sets)

        self.adaption_losses = nn.ModuleList(adaption_losses)  # registers parameters
        self.fuzzy_sets = fuzzy_sets
        self.mean_over_sets = mean_over_sets

        self.add_state("loss", default=torch.tensor(0.0), dist_reduce_fx="sum")
        self.add_state("batch_counter", default=torch.tensor(0), dist_reduce_fx="sum")

    def _check_fuzzy_sets(self, fuzzy_sets: List[Tuple[float, float]]) -> None:
        for i, (left_bound, right_bound) in enumerate(fuzzy_sets):
            if left_bound >= right_bound:
                raise ValueError(
                    f"Fuzzy set {i} with bounds ({left_bound}, {right_bound}) has "
                    f"a left bound greater than or equal to the right bound."
                )

    def update(
        self,
        source: torch.Tensor,
        source_preds: torch.Tensor,
        target: torch.Tensor,
        target_preds: torch.Tensor,
    ) -> None:
        """
        Update the loss with the given data.

        The predictions for the source and target data are checked against the fuzzy
        sets to determine membership.

        Args:
            source: The source features.
            source_preds: The predictions for the source features.
            target: The target features.
            target_preds: The predictions for the target features.
        """
        for fuzzy_set, adaption_loss in zip(self.fuzzy_sets, self.adaption_losses):
            selected_source = source[_membership(source_preds, fuzzy_set)]
            selected_target = target[_membership(target_preds, fuzzy_set)]
            if selected_source.shape[0] > 0 and selected_target.shape[0] > 0:
                self.loss = self.loss + adaption_loss(selected_source, selected_target)
        self.batch_counter += 1

    def compute(self) -> torch.Tensor:
        """
        Compute the loss as either the sum or mean of all subset losses.

        Returns:
            The combined loss.
        """
        if self.batch_counter > 1:
            raise RuntimeError(
                "The update method of this loss was called multiple times before "
                "computing the loss. This is not supported."
            )
        if self.mean_over_sets:
            return self.loss / len(self.adaption_losses)
        else:
            return self.loss


def _membership(preds: torch.Tensor, fuzzy_set: Tuple[float, float]) -> torch.Tensor:
    preds = preds.squeeze() if len(preds.shape) > 1 else preds
    membership = (preds >= fuzzy_set[0]) & (preds < fuzzy_set[1])

    return membership



"""A module for the abstract base class of all approaches."""
import inspect
import warnings
from abc import ABCMeta
from typing import Any, Dict, List, Set

import hydra.utils
import pytorch_lightning as pl
from torch import nn

EXCLUDED_ARGS = ["self", "device", "dtype"]  # init args ignored when checkpointing


class AdaptionApproach(pl.LightningModule, metaclass=ABCMeta):
    """
    This abstract class is the base of all adaption approaches.

    It defines that there needs to be a `feature_extractor`, a `regressor`. These
    members can be accessed via read-only properties. The `feature_extractor` and
    `regressor` are trainable neural networks.

    All child classes are supposed to implement their own constructors. The
    `feature_extractor` and `regressor` should explicitly not be arguments of the
    constructor and should be set by calling [set_model]
    [rul_adapt.approach.abstract.AdaptionApproach.set_model]. This way, the approach can
    be initialized with all hyperparameters first and afterward supplied with the
    networks. This is useful for initializing the networks with pre-trained weights.

    Because models are constructed outside the approach, the default checkpointing
    mechanism of PyTorch Lightning fails to load checkpoints of AdaptionApproaches.
    We extended the checkpointing mechanism by implementing the `on_save_checkpoint`
    and `on_load_checkpoint` callbacks to make it work. If a subclass uses an
    additional model, besides feature extractor and regressor, that is not
    initialized in the constructor, the subclass needs to implement the
    `CHECKPOINT_MODELS` class variable. This variable is a list of model names to be
    included in the checkpoint. For example, if your approach has an additional model
    `self._domain_disc`, the `CHECKPOINT_MODELS` variable should be set to
    `['_domain_disc']`. Otherwise, loading a checkpoint of this approach will fail.
    """

    CHECKPOINT_MODELS: List[str] = []

    _feature_extractor: nn.Module
    _regressor: nn.Module

    _hparams_initial: Dict[str, Any]
    _logged_models: Dict[str, Set[str]]

    def set_model(
        self,
        feature_extractor: nn.Module,
        regressor: nn.Module,
        *args: Any,
        **kwargs: Any,
    ) -> None:
        """
        Set the feature extractor and regressor for this approach.

        Child classes can override this function to add additional models to an
        approach. The `args` and `kwargs` making this possible are ignored in this
        function.

        Args:
            feature_extractor: The feature extraction network.
            regressor: The RUL regression network.
        """
        self._feature_extractor = feature_extractor
        self._regressor = regressor
        if args:
            warnings.warn("Additional position args were supplied, which are ignored.")
        if kwargs:
            warnings.warn("Additional keyword args were supplied, which are ignored.")

        self.log_model_hyperparameters("feature_extractor", "regressor")

    def log_model_hyperparameters(self, *model_names: str) -> None:
        hparams_initial = self.hparams_initial
        if not hasattr(self, "_logged_models"):
            self._logged_models = {}
            hparams_initial["model"] = {}

        for model_name in model_names:
            model_hparams = self._get_model_hparams(model_name)
            hparams_initial["model"].update(model_hparams)
            self._logged_models[model_name] = set(model_hparams.keys())

        self._hparams_initial = hparams_initial
        self._set_hparams(self._hparams_initial)

    def _get_model_hparams(self, model_name):
        prefix = model_name.lstrip("_")
        model = getattr(self, model_name)
        hparams = {prefix: {"type": type(model).__name__}}
        init_args = _get_init_args(model, "logging model hyperparameters")
        hparams[prefix].update(init_args)

        return hparams

    @property
    def feature_extractor(self) -> nn.Module:
        """The feature extraction network."""
        if hasattr(self, "_feature_extractor"):
            return self._feature_extractor
        else:
            raise RuntimeError("Feature extractor used before 'set_model' was called.")

    @property
    def regressor(self) -> nn.Module:
        """The RUL regression network."""
        if hasattr(self, "_regressor"):
            return self._regressor
        else:
            raise RuntimeError("Regressor used before 'set_model' was called.")

    def on_save_checkpoint(self, checkpoint: Dict[str, Any]) -> None:
        del checkpoint["hyper_parameters"]["model"]
        checkpoint["logged_models"] = list(self._logged_models)
        to_checkpoint = ["_feature_extractor", "_regressor"] + self.CHECKPOINT_MODELS
        configs = {m: _get_hydra_config(getattr(self, m)) for m in to_checkpoint}
        checkpoint["model_configs"] = configs

    def on_load_checkpoint(self, checkpoint: Dict[str, Any]) -> None:
        for name, config in checkpoint["model_configs"].items():
            setattr(self, name, hydra.utils.instantiate(config))
        self.log_model_hyperparameters(*checkpoint["logged_models"])


def _get_hydra_config(model: nn.Module) -> Dict[str, Any]:
    model_type = type(model)
    class_name = f"{model_type.__module__}.{model_type.__qualname__}"
    config = {"_target_": class_name, **_get_init_args(model)}

    return config


def _get_init_args(
    obj: nn.Module, activity: str = "writing a checkpoint"
) -> Dict[str, Any]:
    if isinstance(obj, nn.ModuleList):
        # workaround because ModuleList's init arg is shadowed by a property
        init_args = {"modules": [_get_hydra_config(m) for m in obj]}
    elif isinstance(obj, nn.Sequential):
        # workaround because Sequential expects positional args only
        init_args = {"_args_": [_get_hydra_config(m) for m in obj]}
    else:
        signature = inspect.signature(type(obj).__init__)
        init_args = dict()
        arg_names = filter(lambda a: a not in EXCLUDED_ARGS, signature.parameters)
        for arg_name in arg_names:
            _check_has_attr(obj, arg_name, activity)
            arg = getattr(obj, arg_name)
            if isinstance(arg, nn.Module):
                arg = _get_hydra_config(arg)
            init_args[arg_name] = arg

    return init_args


def _check_has_attr(obj: Any, param: str, activity: str) -> None:
    if not hasattr(obj, param):
        raise RuntimeError(
            f"Error while {activity}. "
            f"The nn.Module of type '{type(obj)}' has an initialization parameter "
            f"named '{param}' which is not saved as a member variable, i.e. "
            f"'self.{param}'. Therefore, we cannot retrieve the value of "
            f"'{param}' the object was initialized with."
        )



"""This module contains the necessary neural networks to build a RUL estimator.

In general, a RUL estimator network consists of two parts: a feature extractor and a
regression head. The feature extractor is a network that transforms the input feature
windows into a latent feature vector. The regression head maps these latent features
to a scalar RUL value. The feature extractors can be found in the [cnn]
[rul_adapt.model.cnn] and [rnn][rul_adapt.model.rnn] modules. The regression head in
the [head][rul_adapt.model.head] module.

Some domain adaption approaches use a domain discriminator. The networks in the
[head][rul_adapt.model.head] module can be used to construct them, too.

Examples:
    >>> import torch
    >>> from rul_adapt import model
    >>> feature_extractor = model.CnnExtractor(14,[32, 16],30,fc_units=8)
    >>> regressor = model.FullyConnectedHead(8, [4, 1])
    >>> latent_features = feature_extractor(torch.randn(10, 14, 30))
    >>> latent_features.shape
    torch.Size([10, 8])
    >>> rul = regressor(latent_features)
    >>> rul.shape
    torch.Size([10, 1])
"""

from .cnn import CnnExtractor
from .head import FullyConnectedHead
from .rnn import LstmExtractor, GruExtractor
from .wrapper import ActivationDropoutWrapper
from .two_stage import TwoStageExtractor



import math
from typing import Any, Literal, Optional

import torch
import torchmetrics


class RULScore(torchmetrics.Metric):
    is_differentiable = True
    higher_is_better = None
    full_state_update = False

    loss: torch.Tensor
    total: torch.Tensor

    def __init__(
        self, mode: Literal["phm08", "phm12"] = "phm08", mean: Optional[bool] = None
    ):
        super().__init__()
        if mode == "phm08":
            self.pos_factor = 10.0
            self.neg_factor = -13.0
            self.offset = -1.0
            self.as_percentage = False
            self.mean = False
        elif mode == "phm12":
            self.pos_factor = 20 / math.log(0.5)
            self.neg_factor = -5 / math.log(0.5)
            self.offset = 0.0
            self.as_percentage = True
            self.mean = True
        else:
            raise ValueError(f"Unknown RUL score mode: {mode}")

        if mean is not None:
            self.mean = mean

        self.add_state("loss", torch.tensor(0.0), dist_reduce_fx="sum")
        self.add_state("total", torch.tensor(0.0), dist_reduce_fx="sum")

    def update(self, inputs: torch.Tensor, targets: torch.Tensor) -> None:
        self.loss = self.loss + rul_score(
            inputs,
            targets,
            self.pos_factor,
            self.neg_factor,
            self.offset,
            self.as_percentage,
        )
        self.total = self.total + inputs.shape[0]

    def compute(self) -> Any:
        if self.mean:
            return self.loss / self.total
        else:
            return self.loss


def rul_score(
    inputs: torch.Tensor,
    targets: torch.Tensor,
    pos_factor: float,
    neg_factor: float,
    offset: float,
    as_percentage: bool,
) -> torch.Tensor:
    dist = inputs - targets
    if as_percentage:
        dist = dist / targets * 100
    factors = torch.ones_like(dist)
    factors[dist >= 0] /= pos_factor
    factors[dist < 0] /= neg_factor
    dist = torch.exp(dist * factors) + offset
    score = dist.sum()

    return score



from .cnn_dann import get_cnn_dann, get_cnn_dann_config, cnn_dann_from_config
from .latent_align import (
    get_latent_align,
    get_latent_align_config,
    latent_align_from_config,
)
from .tbigru.functional import get_tbigru, get_tbigru_config, tbigru_from_config
from .adarul.functional import get_adarul, get_adarul_config, adarul_from_config
from .consistency.functional import (
    get_consistency_dann_config,
    get_consistency_dann,
    consistency_dann_from_config,
)
from .lstm_dann.functional import (
    get_lstm_dann,
    get_lstm_dann_config,
    lstm_dann_from_config,
)



from typing import Tuple, Any

import hydra
import omegaconf
import pytorch_lightning as pl
import rul_datasets

import rul_adapt
from rul_adapt.approach import DannApproach


def get_cnn_dann(
    source_fd: int, target_fd: int, **trainer_kwargs: Any
) -> Tuple[rul_datasets.DomainAdaptionDataModule, DannApproach, pl.Trainer]:
    """
    Construct an CNN-DANN approach for CMAPSS with the original hyperparameters.

    The adaption tasks 1-->4, 2-->3, 3-->2 and 4-->1 are missing because they were not
    investigated in the paper.

    Examples:
        ```pycon
        >>> import rul_adapt
        >>> dm, dann, trainer = rul_adapt.construct.get_cnn_dann(3, 1)
        >>> trainer.fit(dann, dm)
        >>> trainer.test(dann, dm)
        ```

    Args:
        source_fd: The source FD of CMAPSS.
        target_fd: The target FD of CMAPSS.
        trainer_kwargs: Overrides for the trainer class.
    Returns:
        dm: The data module for adaption of two CMAPSS sub-datasets.
        dann: The DANN approach with feature extractor, regressor and domain disc.
        trainer: The trainer object.
    """
    config = get_cnn_dann_config(source_fd, target_fd)
    dm, dann, trainer = cnn_dann_from_config(config, **trainer_kwargs)

    return dm, dann, trainer


def get_cnn_dann_config(source_fd: int, target_fd: int) -> omegaconf.DictConfig:
    """
    Get a configuration for the CNN-DANN approach.

    The adaption tasks 1-->4, 2-->3, 3-->2 and 4-->1 are missing because they were
    not investigated in the paper. The configuration can be modified and fed to
    [cnn_dann_from_config][rul_adapt.construct.cnn_dann.cnn_dann_from_config] to
    create the approach.

    Args:
        source_fd: The source FD of CMAPSS.
        target_fd: The target FD of CMAPSS.
    Returns:
        The CNN-DANN configuration.
    """
    _validate(source_fd, target_fd)
    with hydra.initialize("config", version_base="1.1"):
        config = hydra.compose(
            "base",
            overrides=[f"dm.source.fd={source_fd}", f"dm.target.fd={target_fd}"],
        )

    return config


def cnn_dann_from_config(
    config: omegaconf.DictConfig, **trainer_kwargs: Any
) -> Tuple[rul_datasets.DomainAdaptionDataModule, DannApproach, pl.Trainer]:
    """
    Construct a CNN-DANN approach from a configuration.

    The configuration can be created by calling [get_cnn_dann_config]
    [rul_adapt.construct.cnn_dann.get_cnn_dann_config].

    Args:
        config: The CNN-DANN configuration.
        trainer_kwargs: Overrides for the trainer class.
    Returns:
        dm: The data module for adaption of two CMAPSS sub-datasets.
        dann: The DANN approach with feature extractor, regressor and domain disc.
        trainer: The trainer object.
    """
    source = hydra.utils.instantiate(config.dm.source)
    target = source.get_compatible(**config.dm.target)
    dm = rul_datasets.DomainAdaptionDataModule(
        rul_datasets.RulDataModule(source, config.dm.batch_size),
        rul_datasets.RulDataModule(target, config.dm.batch_size),
    )

    feature_extractor = hydra.utils.instantiate(config.feature_extractor)
    regressor = hydra.utils.instantiate(config.regressor)
    domain_disc = hydra.utils.instantiate(config.domain_disc)
    rul_adapt.approach.cnn_dann.init_weights(feature_extractor, regressor)

    dann = hydra.utils.instantiate(config.dann)
    dann.set_model(feature_extractor, regressor, domain_disc)

    trainer = hydra.utils.instantiate(config.trainer, **trainer_kwargs)

    return dm, dann, trainer


def _validate(source_fd: int, target_fd: int) -> None:
    task = (source_fd, target_fd)
    if source_fd == target_fd:
        raise ValueError(
            f"No configuration for adapting from FD{source_fd:03} to itself."
        )
    elif 1 > source_fd or source_fd > 4:
        raise ValueError(f"CMAPSS has only FD001 to FD004 but no FD{source_fd:03}")
    elif 1 > target_fd or target_fd > 4:
        raise ValueError(f"CMAPSS has only FD001 to FD004 but no FD{target_fd:03}")
    elif task == (1, 4) or task == (2, 3) or task == (3, 2) or task == (4, 1):
        raise ValueError(
            f"No configuration for adapting from FD{source_fd:03} to FD{target_fd:03}."
        )



"""A module of feature extractors based on convolutional neural networks."""

from typing import List, Optional, Union, Type

import torch
from torch import nn

from rul_adapt import utils
from rul_adapt.utils import pairwise


class CnnExtractor(nn.Module):
    """A Convolutional Neural Network (CNN) based network that extracts a feature
    vector from same-length time windows.

    This feature extractor consists of multiple CNN layers and an optional fully
    connected (FC) layer. Each CNN layer can be configured with a number of filters
    and a kernel size. Additionally, batch normalization, same-padding and dropout
    can be applied. The fully connected layer can have a separate dropout
    probability.

    Both CNN and FC layers use ReLU activation functions by default. Custom
    activation functions can be set for each layer type.

    The data flow is as follows: `Input --> CNN x n --> [FC] --> Output`

    The expected input shape is `[batch_size, num_features, window_size]`. The output
    of this network is always flattened to `[batch_size, num_extracted_features]`.

    Examples:

        Without FC
        >>> import torch
        >>> from rul_adapt.model import CnnExtractor
        >>> cnn = CnnExtractor(14,units=[16, 1],seq_len=30)
        >>> cnn(torch.randn(10, 14, 30)).shape
        torch.Size([10, 26])

        With FC
        >>> import torch
        >>> from rul_adapt.model import CnnExtractor
        >>> cnn = CnnExtractor(14,units=[16, 1],seq_len=30,fc_units=16)
        >>> cnn(torch.randn(10, 14, 30)).shape
        torch.Size([10, 16])
    """

    def __init__(
        self,
        input_channels: int,
        units: List[int],
        seq_len: int,
        kernel_size: Union[int, List[int]] = 3,
        dilation: int = 1,
        stride: int = 1,
        padding: bool = False,
        fc_units: Optional[int] = None,
        dropout: float = 0.0,
        fc_dropout: float = 0.0,
        batch_norm: bool = False,
        act_func: Type[nn.Module] = nn.ReLU,
        fc_act_func: Type[nn.Module] = nn.ReLU,
    ):
        """
        Create a new CNN-based feature extractor.

        The `units` are the number of output filters for each CNN layer. The
        `seq_len` is needed to calculate the input units for the FC layer. The kernel
        size of each CNN layer can be set by passing a list to `kernel_size`. If an
        integer is passed, each layer has the same kernel size. If `padding` is true,
        same-padding is applied before each CNN layer, which keeps the window_size
        the same. If `batch_norm` is set, batch normalization is applied for each CNN
        layer. If `fc_units` is set, a fully connected layer is appended.

        Dropout can be applied to each CNN layer by setting `conv_dropout` to a
        number greater than zero. The same is valid for the fully connected layer and
        `fc_dropout`. Dropout will never be applied to the input layer.

        The whole network uses ReLU activation functions. This can be customized by
        setting either `conv_act_func` or `fc_act_func`.

        Args:
            input_channels: The number of input channels.
            units: The list of output filters for the CNN layers.
            seq_len: The window_size of the input data.
            kernel_size: The kernel size for the CNN layers. Passing an integer uses
                         the same kernel size for each layer.
            dilation: The dilation for the CNN layers.
            stride: The stride for the CNN layers.
            padding: Whether to apply same-padding before each CNN layer.
            fc_units: Number of output units for the fully connected layer.
            dropout: The dropout probability for the CNN layers.
            fc_dropout: The dropout probability for the fully connected layer.
            batch_norm: Whether to use batch normalization on the CNN layers.
            act_func: The activation function for the CNN layers.
            fc_act_func: The activation function for the fully connected layer.
        """
        super().__init__()

        self.input_channels = input_channels
        self.units = units
        self.seq_len = seq_len
        self.kernel_size = kernel_size
        self.dilation = dilation
        self.stride = stride
        self.padding = padding
        self.fc_units = fc_units
        self.dropout = dropout
        self.fc_dropout = fc_dropout
        self.batch_norm = batch_norm
        self.act_func = utils.str2callable(act_func, restriction="torch.nn")
        self.fc_act_func = utils.str2callable(fc_act_func, restriction="torch.nn")

        self._kernel_sizes = (
            [self.kernel_size] * len(self.units)
            if isinstance(self.kernel_size, int)
            else self.kernel_size
        )
        self._layers = self._get_layers()

    def _get_layers(self) -> nn.Module:
        layers = nn.Sequential()
        filter_iter = pairwise([self.input_channels] + self.units)
        layer_iter = zip(filter_iter, self._kernel_sizes)
        for i, ((in_ch, out_ch), ks) in enumerate(layer_iter):
            layers.add_module(f"conv_{i}", self._get_conv_layer(in_ch, out_ch, ks, i))

        layers.append(nn.Flatten())
        if self.fc_units is not None:
            flat_dim = self._get_flat_dim()
            layers.add_module("fc", self._get_fc_layer(flat_dim, self.fc_units))

        return layers

    def _get_conv_layer(
        self,
        input_channels: int,
        output_channels: int,
        kernel_size: int,
        num_layer: int,
    ) -> nn.Module:
        conv_layer = nn.Sequential()
        if num_layer > 0 and self.dropout > 0:
            conv_layer.append(nn.Dropout1d(self.dropout))
        conv_layer.append(
            nn.Conv1d(
                input_channels,
                output_channels,
                kernel_size,
                dilation=self.dilation,
                stride=self.stride,
                bias=not self.batch_norm,
                padding=self._get_padding(),
            )
        )
        if self.batch_norm:
            conv_layer.append(nn.BatchNorm1d(output_channels))
        conv_layer.append(self.act_func())

        return conv_layer

    def _get_fc_layer(self, input_units: int, output_units: int) -> nn.Module:
        fc_layer = nn.Sequential()
        if self.fc_dropout > 0:
            fc_layer.append(nn.Dropout(self.fc_dropout))
        fc_layer.append(nn.Linear(input_units, output_units))
        fc_layer.append(self.fc_act_func())

        return fc_layer

    def _get_padding(self) -> str:
        return "same" if self.padding else "valid"

    def _get_flat_dim(self) -> int:
        after_conv = self.seq_len
        for kernel_size in self._kernel_sizes:
            cut_off = 0 if self.padding else self.dilation * (kernel_size - 1)
            after_conv = (after_conv - cut_off - 1) // self.stride + 1
        flat_dim = after_conv * self.units[-1]

        return flat_dim

    def forward(self, inputs: torch.Tensor) -> torch.Tensor:
        return self._layers(inputs)



"""The Maximum Mean Discrepancy (MMD) approach uses the distance measure of the same
name to adapt a feature extractor. This implementation uses a multi-kernel variant of
the [MMD loss][rul_adapt.loss.adaption.MaximumMeanDiscrepancyLoss] with bandwidths
set via the median heuristic.

```python
Source --> FeatEx --> Source Feats -----------> Regressor  --> RUL Prediction
        ^         |                 |
        |         |                 v
Target --         --> Target Feats -->  MMD Loss
```

It was first introduced by [Long et al.](
https://dl.acm.org/doi/10.5555/3045118.3045130) as Deep Adaption Network (DAN) for
image classification.

Used In:
    * Cao et al. (2021). **Transfer learning for remaining useful life prediction of
    multi-conditions bearings based on bidirectional-GRU network.**
    *Measurement: Journal of the International Measurement Confederation*, *178*.
    [10.1016/j.measurement.2021.109287](https://doi.org/10.1016/j.measurement.2021.109287)
    * Krokotsch et al. (2020). **A Novel Evaluation Framework for Unsupervised Domain
    Adaption on Remaining Useful Lifetime Estimation.**
    *2020 IEEE International Conference on Prognostics and Health Management (ICPHM)*.
    [10.1109/ICPHM49022.2020.9187058](https://doi.org/10.1109/ICPHM49022.2020.9187058)
"""

from typing import List, Literal, Any

import torch

import rul_adapt
from rul_adapt import utils
from rul_adapt.approach.abstract import AdaptionApproach
from rul_adapt.approach.evaluation import AdaptionEvaluator


class MmdApproach(AdaptionApproach):
    """The MMD uses the Maximum Mean Discrepancy to adapt a feature extractor to
    be used with the source regressor.

    The regressor needs the same number of input units as the feature extractor has
    output units.

    Examples:
        >>> from rul_adapt import model
        >>> from rul_adapt import approach
        >>> feat_ex = model.CnnExtractor(1, [16, 16, 1], 10, fc_units=16)
        >>> reg = model.FullyConnectedHead(16, [1])
        >>> mmd = approach.MmdApproach(0.01)
        >>> mmd.set_model(feat_ex, reg)
    """

    def __init__(
        self,
        mmd_factor: float,
        num_mmd_kernels: int = 5,
        loss_type: Literal["mse", "rmse", "mae"] = "mse",
        rul_score_mode: Literal["phm08", "phm12"] = "phm08",
        evaluate_degraded_only: bool = False,
        **optim_kwargs: Any,
    ) -> None:
        """
        Create a new MMD approach.

        The strength of the influence of the MMD loss on the feature
        extractor is controlled by the `mmd_factor`. The higher it is, the stronger
        the influence.

        For more information about the possible optimizer keyword arguments,
        see [here][rul_adapt.utils.OptimizerFactory].

        Args:
            mmd_factor: The strength of the MMD loss' influence.
            num_mmd_kernels: The number of kernels for the MMD loss.
            loss_type: The type of regression loss, either 'mse', 'rmse' or 'mae'.
            rul_score_mode: The mode for the val and test RUL score, either 'phm08'
                            or 'phm12'.
            evaluate_degraded_only: Whether to only evaluate the RUL score on degraded
                                    samples.
            **optim_kwargs: Keyword arguments for the optimizer, e.g. learning rate.
        """
        super().__init__()

        self.mmd_factor = mmd_factor
        self.num_mmd_kernels = num_mmd_kernels
        self.loss_type = loss_type
        self.rul_score_mode = rul_score_mode
        self.evaluate_degraded_only = evaluate_degraded_only
        self.optim_kwargs = optim_kwargs

        # training metrics
        self.train_source_loss = utils.get_loss(self.loss_type)
        self.mmd_loss = rul_adapt.loss.MaximumMeanDiscrepancyLoss(self.num_mmd_kernels)
        self._get_optimizer = utils.OptimizerFactory(**self.optim_kwargs)
        self.evaluator = AdaptionEvaluator(
            self.forward, self.log, self.rul_score_mode, self.evaluate_degraded_only
        )

        self.save_hyperparameters()

    def configure_optimizers(self) -> utils.OptimizerLRSchedulerConfig:
        """Configure an optimizer."""
        return self._get_optimizer(self.parameters())

    def forward(self, inputs: torch.Tensor) -> torch.Tensor:
        """Predict the RUL values for a batch of input features."""
        return self.regressor(self.feature_extractor(inputs))

    def training_step(self, batch: List[torch.Tensor], batch_idx: int) -> torch.Tensor:
        """
        Execute one training step.

        The `batch` argument is a list of three tensors representing the source
        features, source labels and target features. Both types of features are fed
        to the feature extractor. Then the regression loss for the source domain and
        the MMD loss between domains is computed. The regression, MMD and combined
        loss are logged.

        Args:
            batch: A list of a source feature, source label and target feature tensors.
            batch_idx: The index of the current batch.
        Returns:
            The combined loss.
        """
        source, source_labels, target = batch
        source_labels = source_labels[:, None]

        source = self.feature_extractor(source)
        target = self.feature_extractor(target)
        rul_preds = self.regressor(source)

        mse_loss = self.train_source_loss(rul_preds, source_labels)
        mmd_loss = self.mmd_loss(source, target)
        loss = mse_loss + self.mmd_factor * mmd_loss

        self.log("train/loss", loss)
        self.log("train/source_loss", self.train_source_loss)
        self.log("train/mmd", self.mmd_loss)

        return loss

    def validation_step(
        self, batch: List[torch.Tensor], batch_idx: int, dataloader_idx: int
    ) -> None:
        """
        Execute one validation step.

        The `batch` argument is a list of two tensors representing features and
        labels. A RUL prediction is made from the features and the validation RMSE
        and RUL score are calculated. The metrics recorded for dataloader idx zero
        are assumed to be from the source domain and for dataloader idx one from the
        target domain. The metrics are written to the configured logger under the
        prefix `val`.

        Args:
            batch: A list containing a feature and a label tensor.
            batch_idx: The index of the current batch.
            dataloader_idx: The index of the current dataloader (0: source, 1: target).
        """
        domain = utils.dataloader2domain(dataloader_idx)
        self.evaluator.validation(batch, domain)

    def test_step(
        self, batch: List[torch.Tensor], batch_idx: int, dataloader_idx: int
    ) -> None:
        """
        Execute one test step.

        The `batch` argument is a list of two tensors representing features and
        labels. A RUL prediction is made from the features and the validation RMSE
        and RUL score are calculated. The metrics recorded for dataloader idx zero
        are assumed to be from the source domain and for dataloader idx one from the
        target domain. The metrics are written to the configured logger under the
        prefix `test`.

        Args:
            batch: A list containing a feature and a label tensor.
            batch_idx: The index of the current batch.
            dataloader_idx: The index of the current dataloader (0: source, 1: target).
        """
        domain = utils.dataloader2domain(dataloader_idx)
        self.evaluator.test(batch, domain)



"""The latent space alignment approach uses several auxiliary losses to align the
latent space of the source and target domain produced by a shared feature extractor:

* **Healthy State Alignment:** Pushes the healthy data of both domains into a single
  compact cluster
* **Degradation Direction Alignment:** Minimizes the angle between degraded data
  points with the healthy cluster as origin
* **Degradation Level Alignment:** Aligns the distance of degraded data points from the
  healthy cluster to the number of time steps in degradation
* **Degradation Fusion:** Uses a
  [MMD loss][rul_adapt.loss.adaption.MaximumMeanDiscrepancyLoss] to align the
  distribution of both domains

Which features are considered in the healthy state and which in degradation is either
determined by taking the first few steps of each time series or by using a
first-time-to-predict estimation. The first variant is used for CMAPSS, the second
for XJTU-SY.

The approach was introduced by [Zhang et al.](
https://doi.org/10.1016/j.ress.2021.107556) in 2021. For applying the approach on raw
vibration data, i.e. XJTU-SY, it uses a [windowing scheme]
[rul_adapt.approach.latent_align.extract_chunk_windows] and
[first-point-to-predict estimation]
[rul_adapt.approach.latent_align.LatentAlignFttpApproach] introduced by [Li et al.](
https://doi.org/10.1016/j.knosys.2020.105843) in 2020."""

from typing import Tuple, List, Any, Optional, Literal

import numpy as np
import torch
from rul_datasets.utils import feature_to_tensor
from torch import nn

import rul_adapt
from rul_adapt import utils
from rul_adapt.approach.abstract import AdaptionApproach
from rul_adapt.approach.evaluation import AdaptionEvaluator
from rul_adapt.utils import OptimizerLRSchedulerConfig


class LatentAlignFttpApproach(AdaptionApproach):
    """This first-point-to-predict estimation approach trains a GAN on healthy state
    bearing data. The discriminator can be used afterward to compute a health
    indicator for each bearing.

    The feature extractor and regressor models are used as the discriminator. The
    regressor is not allowed to have an activation function on its last layer and
    needs to use only a single output neuron because [BCEWithLogitsLoss]
    [torch.nn.BCEWithLogitsLoss] is used. The generator receives noise with the shape
    [batch_size, 1, noise_dim]. The generator needs an output with enough elements so
    that it can be reshaped to the same shape as the real input data. The reshaping
    is done internally.

    Both generator and discriminator are trained at once by using a
    [Gradient Reversal Layer][rul_adapt.loss.adaption.GradientReversalLayer]
    between them.

    Examples:
        >>> from rul_adapt import model, approach
        >>> feat_ex = model.CnnExtractor(1, [16, 16, 1], 10, fc_units=16)
        >>> reg = model.FullyConnectedHead(16, [1])
        >>> gen = model.CnnExtractor(1, [1], 10, padding=True)
        >>> fttp_model = approach.LatentAlignFttpApproach(1e-4, 10)
        >>> fttp_model.set_model(feat_ex, reg, gen)
        >>> health_indicator = fttp_model(torch.randn(16, 1, 10)).std()

    """

    CHECKPOINT_MODELS = ["_generator"]

    _generator: nn.Module

    def __init__(
        self,
        noise_dim: int,
        **optim_kwargs: Any,
    ):
        """
        Create a new FTTP estimation approach.

        The generator is set by the `set_model` function together with the feature
        extractor and regressor.

        For more information about the possible optimizer keyword arguments,
        see [here][rul_adapt.utils.OptimizerFactory].

        Args:
            noise_dim: The size of the last dimension of the noise tensor.
            **optim_kwargs: Keyword arguments for the optimizer, e.g. learning rate.
        """
        super().__init__()

        self.noise_dim = noise_dim
        self.optim_kwargs = optim_kwargs

        self.gan_loss = torch.nn.BCEWithLogitsLoss()
        self.grl = rul_adapt.loss.adaption.GradientReversalLayer()
        self._get_optimizer = utils.OptimizerFactory(**self.optim_kwargs)

        self.save_hyperparameters()

    def set_model(
        self,
        feature_extractor: nn.Module,
        regressor: nn.Module,
        generator: Optional[nn.Module] = None,
        *args: Any,
        **kwargs: Any,
    ) -> None:
        """
        Set the feature extractor, regressor (forming the discriminator) and
        generator for this approach.

        The regressor is not allowed to have an activation function on its last layer
        and needs to use only a single output neuron. The generator receives noise
        with the shape [batch_size, 1, noise_dim]. The generator needs an output with
        enough elements so that it can be reshaped to the same shape as the real
        input data. The reshaping is done internally.

        Args:
            feature_extractor: The feature extraction network.
            regressor: The regressor functioning as the head of the discriminator.
            generator: The generator network.
        """
        super().set_model(feature_extractor, regressor)
        if generator is None:
            raise ValueError("Generator not set. This approach is unlikely to work.")
        self._generator = generator

    @property
    def generator(self):
        """The generator network."""
        if hasattr(self, "_generator"):
            return self._generator
        else:
            raise RuntimeError("Generator used before 'set_model' was called.")

    def configure_optimizers(self) -> OptimizerLRSchedulerConfig:
        """Configure an optimizer for the generator and discriminator."""
        return self._get_optimizer(self.parameters())

    def forward(self, inputs: torch.Tensor) -> torch.Tensor:
        """Predict the health indicator for the given inputs."""
        return self.regressor(self.feature_extractor(inputs))

    def training_step(self, batch: Tuple[torch.Tensor, torch.Tensor]) -> torch.Tensor:
        """
        Execute one training step.

        The batch is a tuple of the features and the labels. The labels are ignored.
        A noise tensor is passed to the generator to generate fake features. The
        discriminator classifies if the features are real or fake and the binary
        cross entropy loss is calculated. Real features receive the label zero and
        the fake features one.

        Both generator and discriminator are trained at once by using a
        [Gradient Reversal Layer][rul_adapt.loss.adaption.GradientReversalLayer]
        between them. At the end, the loss is logged.

        Args:
            batch: A tuple of feature and label tensors.

        Returns:
            The classification loss.
        """
        features, _ = batch
        batch_size = features.shape[0]
        device: torch.device = self.device  # type: ignore[assignment]

        pred_real = self.forward(features)
        loss_real = self.gan_loss(pred_real, torch.zeros(batch_size, 1, device=device))

        noise = torch.randn(batch_size, 1, self.noise_dim, device=device)
        fake_features = self.grl(self.generator(noise)).reshape_as(features)
        pred_fake = self.forward(fake_features)
        loss_fake = self.gan_loss(pred_fake, torch.ones(batch_size, 1, device=device))

        loss = (loss_real + loss_fake) / 2
        self.log("train/loss", loss)

        return loss


def get_first_time_to_predict(
    fttp_model: LatentAlignFttpApproach,
    features: np.ndarray,
    window_size: int,
    chunk_size: int,
    healthy_index: int,
    threshold_coefficient: float,
) -> int:
    """
    Get the first time step to predict for the given features.

    The features are pre-processed via the [extract_chunk_windows]
    [rul_adapt.approach.latent_align.extract_chunk_windows] function and fed in
    batches to the `fttp_model`. Each batch consists of the chunk windows that end in
    the same original feature window. The health indicator for the original window is
    calculated as the standard deviation of the predictions of the `fttp_model`.

    The first-time-to-predict is the first time step where the health indicator is
    larger than `threshold_coefficient` times the mean of the health indicator for
    the first `healthy_index` time steps. If the threshold is never exceeded,
    a RuntimeError is raised.

    Args:
        fttp_model: The model to use for the health indicator calculation.
        features: The features to calculate the first-time-to-predict for.
        window_size: The size of the chunk windows to extract.
        chunk_size: The size of the chunks for each chunk window to extract.
        healthy_index: The index of the last healthy time step.
        threshold_coefficient: The threshold coefficient for the health indicator.

    Returns:
        The original window index of the first-time-to-predict.
    """
    if threshold_coefficient <= 1:
        raise ValueError("Threshold coefficient needs to be greater than one.")

    health_indicator = get_health_indicator(
        fttp_model, features, window_size, chunk_size
    )
    offset = len(features) - len(health_indicator)  # windowing cuts off first windows
    healthy = np.mean(health_indicator[: healthy_index - offset])
    over_thresh = np.argwhere(health_indicator > threshold_coefficient * healthy)

    if len(over_thresh) == 0:
        raise RuntimeError("Health indicator never passes threshold.")
    fttp = over_thresh[0, 0] + offset

    return fttp


@torch.no_grad()
def get_health_indicator(
    fttp_model: nn.Module, features: np.ndarray, window_size: int, chunk_size: int
) -> np.ndarray:
    """
    Get the health indicator for the given features.

    The features are pre-processed via the [extract_chunk_windows]
    [rul_adapt.approach.latent_align.extract_chunk_windows] function and fed in
    batches to the `fttp_model`. Each batch consists of the chunk windows that end in
    the same original feature window. The health indicator for the original window is
    calculated as the standard deviation of the predictions of the `fttp_model`.

    The length of the returned health indicator array is shorter than the `features`
    array by `window_size - 1`, due to the chunk windowing. This means the first
    health indicator value belongs to the original window with the index
    `window_size - 1`.

    Args:
        fttp_model: The model to use for the health indicator calculation.
        features: The features to calculate the health indicator for.
        window_size: The size of the chunk windows to extract.
        chunk_size: The size of the chunks for each chunk window to extract.

    Returns:
        The health indicator for the original windows.
    """
    chunked = extract_chunk_windows(features, window_size, chunk_size)
    chunks_per_window = features.shape[1] // chunk_size
    batches = np.split(chunked, len(chunked) // chunks_per_window)
    health_indicator = np.empty(len(batches))
    for i, batch in enumerate(batches):
        preds = fttp_model(feature_to_tensor(batch, torch.float))
        health_indicator[i] = np.std(preds.detach().numpy())

    return health_indicator


def extract_chunk_windows(
    features: np.ndarray, window_size: int, chunk_size: int
) -> np.ndarray:
    """
    Extract chunk windows from the given features of shape `[num_org_windows,
    org_window_size, num_features]`.

    A chunk window is a window that consists of `window_size` chunks. Each original
    window is split into chunks of size `chunk_size`. A chunk window is then formed
    by concatenating chunks from the same position inside `window_size` consecutive
    original windows. Therefore, each original window is represented by
    `org_window_size // chunk_size` chunk windows. The original window size must
    therefor be divisible by the chunk size.

    Args:
        features: The features to extract the chunk windows from.
        window_size: The number of consecutive original windows to form a chunk window
                     from.
        chunk_size: The size of the chunks to extract from the original windows.

    Returns:
        Chunk windows of shape `[num_windows, window_size * chunk_size, num_features]`.
    """
    old_window_size = features.shape[1]
    window_multiplier = old_window_size // chunk_size
    num_new_windows = (features.shape[0] - window_size + 1) * window_multiplier

    chunk_idx = np.tile(np.arange(chunk_size), num_new_windows * window_size)
    intra_offsets = np.tile(np.arange(window_size), num_new_windows) * old_window_size
    inter_offsets = np.repeat(np.arange(num_new_windows), window_size) * chunk_size
    offsets = np.repeat(intra_offsets + inter_offsets, chunk_size)
    window_idx = chunk_idx + offsets

    flat_features = features.reshape((-1, features.shape[2]))
    flat_windows = flat_features[window_idx]
    windows = flat_windows.reshape((num_new_windows, window_size * chunk_size, -1))

    return windows


class ChunkWindowExtractor:
    def __init__(self, window_size: int, chunk_size: int) -> None:
        self.window_size = window_size
        self.chunk_size = chunk_size

    def __call__(
        self, features: np.ndarray, targets: np.ndarray
    ) -> Tuple[np.ndarray, np.ndarray]:
        chunks_per_window = features.shape[1] // self.chunk_size
        chunked = extract_chunk_windows(features, self.window_size, self.chunk_size)
        targets = targets[self.window_size - 1 :].repeat(chunks_per_window)

        return chunked, targets


class LatentAlignApproach(AdaptionApproach):
    """
    The latent alignment approach introduces four latent space alignment losses to
    align the latent space of a shared feature extractor to both source and target
    domain.

    Examples:
        >>> from rul_adapt import model, approach
        >>> feat_ex = model.CnnExtractor(1, [16, 16, 1], 10, fc_units=16)
        >>> reg = model.FullyConnectedHead(16, [1])
        >>> latent_align = approach.LatentAlignApproach(0.1, 0.1, 0.1, 0.1, lr=0.001)
        >>> latent_align.set_model(feat_ex, reg)
    """

    def __init__(
        self,
        alpha_healthy: float,
        alpha_direction: float,
        alpha_level: float,
        alpha_fusion: float,
        loss_type: Literal["mse", "mae", "rmse"] = "mse",
        rul_score_mode: Literal["phm08", "phm12"] = "phm08",
        evaluate_degraded_only: bool = False,
        labels_as_percentage: bool = False,
        **optim_kwargs: Any,
    ) -> None:
        """
        Create a new latent alignment approach.

        Each of the alphas controls the influence of the respective loss on the
        training. Commonly they are all set to the same value.

        For more information about the possible optimizer keyword arguments,
        see [here][rul_adapt.utils.OptimizerFactory].

        Args:
            alpha_healthy: The influence of the healthy state alignment loss.
            alpha_direction: The influence of the degradation direction alignment loss.
            alpha_level: The influence of the degradation level regularization loss.
            alpha_fusion: The influence of the degradation fusion (MMD) loss.
            loss_type: The type of regression loss to use.
            rul_score_mode: The mode for the val and test RUL score, either 'phm08'
                            or 'phm12'.
            evaluate_degraded_only: Whether to only evaluate the RUL score on degraded
                                    samples.
            labels_as_percentage: Whether to multiply labels by 100 to get percentages
            **optim_kwargs: Keyword arguments for the optimizer, e.g. learning rate.
        """
        super().__init__()

        self.alpha_healthy = alpha_healthy
        self.alpha_direction = alpha_direction
        self.alpha_level = alpha_level
        self.alpha_fusion = alpha_fusion
        self.loss_type = loss_type
        self.rul_score_mode = rul_score_mode
        self.evaluate_degraded_only = evaluate_degraded_only
        self.labels_as_percentage = labels_as_percentage
        self.optim_kwargs = optim_kwargs

        # training metrics
        self.train_mse = utils.get_loss(self.loss_type)
        self.healthy_align = rul_adapt.loss.HealthyStateAlignmentLoss()
        self.direction_align = rul_adapt.loss.DegradationDirectionAlignmentLoss()
        self.level_align = rul_adapt.loss.DegradationLevelRegularizationLoss()
        self.fusion_align = rul_adapt.loss.MaximumMeanDiscrepancyLoss(num_kernels=5)
        self._get_optimizer = utils.OptimizerFactory(**self.optim_kwargs)

        self.evaluator = AdaptionEvaluator(
            self.forward, self.log, self.rul_score_mode, self.evaluate_degraded_only
        )

        self.save_hyperparameters()

    def configure_optimizers(self) -> utils.OptimizerLRSchedulerConfig:
        """Configure an optimizer."""
        optim = self._get_optimizer(self.parameters())

        return optim

    def forward(self, features: torch.Tensor) -> torch.Tensor:
        """Predict the RUL values for a batch of input features."""
        preds = self.regressor(self.feature_extractor(features))
        if self.labels_as_percentage:
            preds = self._from_percentage(preds)

        return preds

    def training_step(
        self, batch: Tuple[torch.Tensor, ...], batch_idx: int
    ) -> torch.Tensor:
        """
        Execute one training step.

        The `batch` contains the following tensors in order:

        * The source domain features.
        * The steps in degradation for the source features.
        * The RUL labels for the source features.
        * The target domain features.
        * The steps in degradation for the target features.
        * The healthy state features for both domains.

        The easies way to produce such a batch is using the [LatentAlignDataModule]
        [rul_datasets.adaption.LatentAlignDataModule].

        The source, target and healthy features are passed through the feature
        extractor. Afterward, these high-level features are used to compute the
        alignment losses. The source domain RUL predictions are computed using the
        regressor and used to calculate the MSE loss. The losses are then combined.
        Each separate and the combined loss are logged.

        Args:
            batch: The batch of data.
            batch_idx: The index of the batch.

        Returns:
            The combined loss.
        """
        source, source_degradation_steps, source_labels, *_ = batch
        *_, target, target_degradation_steps, healthy = batch
        if self.labels_as_percentage:
            source_labels = self._to_percentage(source_labels)
        else:
            source_labels = source_labels[:, None]

        healthy = self.feature_extractor(healthy)
        source = self.feature_extractor(source)
        target = self.feature_extractor(target)

        rul_predictions = self.regressor(source)
        mse_loss = self.train_mse(rul_predictions, source_labels)

        healthy_loss = self.healthy_align(healthy)
        direction_loss = self.direction_align(healthy, torch.cat([source, target]))
        level_loss = self.level_align(
            healthy, source, source_degradation_steps, target, target_degradation_steps
        )
        fusion_loss = self.fusion_align(source, target)

        loss = (
            mse_loss
            + self.alpha_healthy * healthy_loss
            + self.alpha_direction * direction_loss
            + self.alpha_level * level_loss
            + self.alpha_fusion * fusion_loss
        )

        self.log("train/loss", loss)
        self.log("train/mse", self.train_mse)
        self.log("train/healthy_align", self.healthy_align)
        self.log("train/direction_align", self.direction_align)
        self.log("train/level_align", self.level_align)
        self.log("train/fusion_align", self.fusion_align)

        return loss

    def _to_percentage(self, source_labels):
        """Convert RUL labels to percentages assuming they are normed between [0, 1]."""
        return source_labels[:, None] * 100

    def _from_percentage(self, predictions: torch.Tensor) -> torch.Tensor:
        return predictions / 100

    def validation_step(
        self, batch: List[torch.Tensor], batch_idx: int, dataloader_idx: int
    ) -> None:
        """
        Execute one validation step.

        The `batch` argument is a list of two tensors representing features and
        labels. A RUL prediction is made from the features and the validation RMSE
        and RUL score are calculated. The metrics recorded for dataloader idx zero
        are assumed to be from the source domain and for dataloader idx one from the
        target domain. The metrics are written to the configured logger under the
        prefix `val`.

        Args:
            batch: A list containing a feature and a label tensor.
            batch_idx: The index of the current batch.
            dataloader_idx: The index of the current dataloader (0: source, 1: target).
        """
        domain = utils.dataloader2domain(dataloader_idx)
        self.evaluator.validation(batch, domain)

    def test_step(
        self, batch: List[torch.Tensor], batch_idx: int, dataloader_idx: int
    ) -> None:
        """
        Execute one test step.

        The `batch` argument is a list of two tensors representing features and
        labels. A RUL prediction is made from the features and the validation RMSE
        and RUL score are calculated. The metrics recorded for dataloader idx zero
        are assumed to be from the source domain and for dataloader idx one from the
        target domain. The metrics are written to the configured logger under the
        prefix `test`.

        Args:
            batch: A list containing a feature and a label tensor.
            batch_idx: The index of the current batch.
            dataloader_idx: The index of the current dataloader (0: source, 1: target).
        """
        domain = utils.dataloader2domain(dataloader_idx)
        self.evaluator.test(batch, domain)



"""The Domain Adversarial Neural Network (DANN) approach uses a domain discriminator
trained on distinguishing the source and target features produced by a shared feature
extractor. A [Gradient Reversal Layer][rul_adapt.loss.adaption.GradientReversalLayer]
(GRL) is used to train the feature extractor on making its source and target outputs
indistinguishable.

```python
Source --> FeatEx --> Source Feats -----------> Regressor  --> RUL Prediction
        ^         |                 |
        |         |                 v
Target --         --> Target Feats -->  GRL --> DomainDisc --> Domain Prediction
```

It was originally introduced by [Ganin et al.](http://jmlr.org/papers/v17/15-239.html)
for image classification.

Used In:
    * da Costa et al. (2020). **Remaining useful lifetime prediction via deep domain
    adaptation.**
    *Reliability Engineering & System Safety*, *195*, 106682.
    [10.1016/J.RESS.2019.106682](https://doi.org/10.1016/J.RESS.2019.106682)
    * Krokotsch et al. (2020). **A Novel Evaluation Framework for Unsupervised Domain
    Adaption on Remaining Useful Lifetime Estimation.**
    *2020 IEEE International Conference on Prognostics and Health Management (ICPHM)*.
    [10.1109/ICPHM49022.2020.9187058](https://doi.org/10.1109/ICPHM49022.2020.9187058)
"""

from typing import Any, Optional, Literal, List

import torch
from torch import nn

import rul_adapt.loss
from rul_adapt import utils
from rul_adapt.approach.abstract import AdaptionApproach
from rul_adapt.approach.evaluation import AdaptionEvaluator
from rul_adapt.model import FullyConnectedHead


class DannApproach(AdaptionApproach):
    """The DANN approach introduces a domain discriminator that is trained on
    distinguishing source and target features as a binary classification problem. The
    features are produced by a shared feature extractor. The loss in the domain
    discriminator is binary cross-entropy.

    The regressor and domain discriminator need the same number of input units as the
    feature extractor has output units. The discriminator is not allowed to have an
    activation function on its last layer and needs to use only a single output
    neuron because [BCEWithLogitsLoss][torch.nn.BCEWithLogitsLoss] is used.

    Examples:
        >>> from rul_adapt import model
        >>> from rul_adapt import approach
        >>> feat_ex = model.CnnExtractor(1, [16, 16, 1], 10, fc_units=16)
        >>> reg = model.FullyConnectedHead(16, [1])
        >>> disc = model.FullyConnectedHead(16, [8, 1], act_func_on_last_layer=False)
        >>> dann = approach.DannApproach(1.0)
        >>> dann.set_model(feat_ex, reg, disc)
    """

    CHECKPOINT_MODELS = ["dann_loss"]

    dann_loss: rul_adapt.loss.DomainAdversarialLoss

    def __init__(
        self,
        dann_factor: float,
        loss_type: Literal["mae", "mse", "rmse"] = "mae",
        rul_score_mode: Literal["phm08", "phm12"] = "phm08",
        evaluate_degraded_only: bool = False,
        **optim_kwargs: Any,
    ):
        """
        Create a new DANN approach.

        The strength of the domain discriminator's influence on the feature
        extractor is controlled by the `dann_factor`. The higher it is, the stronger
        the influence.

        Possible options for the regression loss are `mae`, `mse` and `rmse`.

        The domain discriminator is set by the `set_model` function together with the
        feature extractor and regressor. For more information, see the [approach]
        [rul_adapt.approach] module page.

        For more information about the possible optimizer keyword arguments,
        see [here][rul_adapt.utils.OptimizerFactory].

        Args:
            dann_factor: Strength of the domain DANN loss.
            loss_type: Type of regression loss.
            rul_score_mode: The mode for the val and test RUL score, either 'phm08'
                            or 'phm12'.
            evaluate_degraded_only: Whether to only evaluate the RUL score on degraded
                                    samples.
            **optim_kwargs: Keyword arguments for the optimizer, e.g. learning rate.
        """
        super().__init__()

        self.dann_factor = dann_factor
        self.loss_type = loss_type
        self.rul_score_mode = rul_score_mode
        self.evaluate_degraded_only = evaluate_degraded_only
        self.optim_kwargs = optim_kwargs

        self._get_optimizer = utils.OptimizerFactory(**self.optim_kwargs)

        self.train_source_loss = utils.get_loss(self.loss_type)
        self.evaluator = AdaptionEvaluator(
            self.forward, self.log, self.rul_score_mode, self.evaluate_degraded_only
        )

        self.save_hyperparameters()

    def set_model(
        self,
        feature_extractor: nn.Module,
        regressor: nn.Module,
        domain_disc: Optional[nn.Module] = None,
        *args: Any,
        **kwargs: Any,
    ) -> None:
        """
        Set the feature extractor, regressor, and domain discriminator for this
        approach.

        The discriminator is not allowed to have an activation function on its last
        layer and needs to use only a single output neuron. It is wrapped by a
        [DomainAdversarialLoss][rul_adapt.loss.DomainAdversarialLoss].

        Args:
            feature_extractor: The feature extraction network.
            regressor: The RUL regression network.
            domain_disc: The domain discriminator network.
        """
        domain_disc = self._check_domain_disc(domain_disc)
        super().set_model(feature_extractor, regressor, *args, **kwargs)
        self.dann_loss = rul_adapt.loss.DomainAdversarialLoss(domain_disc)
        self.log_model_hyperparameters("domain_disc")

    def _check_domain_disc(self, domain_disc: Optional[nn.Module]) -> nn.Module:
        if domain_disc is None:
            raise ValueError(
                "No domain discriminator was set. This approach is unlikely to work."
            )
        if (
            isinstance(domain_disc, FullyConnectedHead)
            and domain_disc.act_func_on_last_layer
        ):
            raise ValueError(
                "Domain discriminator has an activation function on its last layer. "
                "This is not allowed due to torch.nn.BCEWithLogitsLoss being used as "
                "its loss. Please set 'act_func_on_last_layer' to False."
            )

        return domain_disc

    @property
    def domain_disc(self):
        """The domain discriminator network."""
        if hasattr(self, "dann_loss"):
            return self.dann_loss.domain_disc
        else:
            raise RuntimeError("Domain disc used before 'set_model' was called.")

    def configure_optimizers(self) -> utils.OptimizerLRSchedulerConfig:
        """Configure an optimizer for the whole model."""
        return self._get_optimizer(self.parameters())

    def forward(self, inputs: torch.Tensor) -> torch.Tensor:
        """Predict the RUL values for a batch of input features."""
        return self.regressor(self.feature_extractor(inputs))

    def training_step(self, batch: List[torch.Tensor], batch_idx: int) -> torch.Tensor:
        """
        Execute one training step.

        The `batch` argument is a list of three tensors representing the source
        features, source labels and target features. Both types of features are fed
        to the feature extractor. Then the regression loss for the source domain and
        the DANN loss between domains is computed. The regression, DANN and combined
        loss are logged.

        Args:
            batch: A list of a source feature, source label and target feature tensors.
            batch_idx: The index of the current batch.
        Returns:
            The combined loss.
        """
        source, source_labels, target = batch
        source_labels = source_labels[:, None]

        source = self.feature_extractor(source)
        target = self.feature_extractor(target)

        rul_preds = self.regressor(source)
        mse_loss = self.train_source_loss(rul_preds, source_labels)

        dann_loss = self.dann_loss(source, target)

        loss = mse_loss + self.dann_factor * dann_loss

        self.log("train/loss", loss)
        self.log("train/source_loss", self.train_source_loss)
        self.log("train/dann", self.dann_loss)

        return loss

    def validation_step(
        self, batch: List[torch.Tensor], batch_idx: int, dataloader_idx: int
    ) -> None:
        """
        Execute one validation step.

        The `batch` argument is a list of two tensors representing features and
        labels. A RUL prediction is made from the features and the validation RMSE
        and RUL score are calculated. The metrics recorded for dataloader idx zero
        are assumed to be from the source domain and for dataloader idx one from the
        target domain. The metrics are written to the configured logger under the
        prefix `val`.

        Args:
            batch: A list containing a feature and a label tensor.
            batch_idx: The index of the current batch.
            dataloader_idx: The index of the current dataloader (0: source, 1: target).
        """
        domain = utils.dataloader2domain(dataloader_idx)
        self.evaluator.validation(batch, domain)

    def test_step(
        self, batch: List[torch.Tensor], batch_idx: int, dataloader_idx: int
    ) -> None:
        """
        Execute one test step.

        The `batch` argument is a list of two tensors representing features and
        labels. A RUL prediction is made from the features and the validation RMSE
        and RUL score are calculated. The metrics recorded for dataloader idx zero
        are assumed to be from the source domain and for dataloader idx one from the
        target domain. The metrics are written to the configured logger under the
        prefix `test`.

        Args:
            batch: A list containing a feature and a label tensor.
            batch_idx: The index of the current batch.
            dataloader_idx: The index of the current dataloader (0: source, 1: target).
        """
        domain = utils.dataloader2domain(dataloader_idx)
        self.evaluator.test(batch, domain)



"""A module for unsupervised domain adaption losses."""

from typing import List, Any, Tuple

import torch
import torchmetrics
from torch import nn

from rul_adapt.loss.utils import calc_pairwise_euclidean, weighted_mean


class MaximumMeanDiscrepancyLoss(torchmetrics.Metric):
    """The Maximum Mean Discrepancy Loss (MMD) is a distance measure between two
    arbitrary distributions.

    The distance is defined as the dot product in a reproducing Hilbert kernel space
    (RHKS) and is zero if and only if the distributions are identical. The RHKS is
    the space of the linear combination of multiple Gaussian kernels with bandwidths
    derived by the median heuristic.

    The source and target feature batches are treated as samples from their
    respective distribution. The linear pairwise distances between the two batches
    are transformed into distances in the RHKS via the kernel trick:

    ```python
    rhks(x, y) = dot(to_rhks(x), to_rhks(y)) = multi_kernel(dot(x, y))
    multi_kernel(distance) = mean([gaussian(distance, bw) for bw in bandwidths])
    gaussian(distance, bandwidth) = exp(-distance * bandwidth)
    ```

    The n kernels will use bandwidths between `median / (2**(n/ 2))` and `median * (
    2**(n / 2))`, where `median` is the median of the linear distances.

    The MMD loss is then calculated as:

    ```python
    mean(rhks(source, source) + rhks(target, target) - 2 * rhks(source, target))
    ```

    This version of MMD is biased, which is acceptable for training purposes.
    """

    is_differentiable = True
    higher_is_better = False
    full_state_update = False

    s2s: torch.Tensor
    t2t: torch.Tensor
    s2t: torch.Tensor

    s2s_total: torch.Tensor
    t2t_total: torch.Tensor
    s2t_total: torch.Tensor

    def __init__(self, num_kernels: int) -> None:
        """
        Create a new MMD loss module with `n` kernels.

        The bandwidths of the Gaussian kernels are derived by the median heuristic.

        Args:
            num_kernels: Number of Gaussian kernels to use.
        """
        super().__init__()

        self.num_kernels = num_kernels

        self.add_state("s2s", torch.tensor(0.0), dist_reduce_fx="sum")
        self.add_state("t2t", torch.tensor(0.0), dist_reduce_fx="sum")
        self.add_state("s2t", torch.tensor(0.0), dist_reduce_fx="sum")

        self.add_state("s2s_total", torch.tensor(0.0), dist_reduce_fx="sum")
        self.add_state("t2t_total", torch.tensor(0.0), dist_reduce_fx="sum")
        self.add_state("s2t_total", torch.tensor(0.0), dist_reduce_fx="sum")

    def update(
        self, source_features: torch.Tensor, target_features: torch.Tensor
    ) -> None:
        """
        Compute the MMD loss between source and target feature distributions.

        Args:
            source_features: Source features with shape `[batch_size, num_features]`
            target_features: Target features with shape `[batch_size, num_features]`

        Returns:
            scalar MMD loss
        """
        feats = torch.cat([source_features, target_features], dim=0)
        distances = calc_pairwise_euclidean(feats, feats)

        gammas = _get_gammas(distances, self.num_kernels)
        distances = _calc_multi_kernel(distances, gammas)

        batch_size = source_features.shape[0]
        s2s, t2t, s2t = _split_distances(distances, batch_size)

        self.s2s = self.s2s + s2s.sum()
        self.t2t = self.t2t + t2t.sum()
        self.s2t = self.s2t + s2t.sum()

        self.s2s_total = self.s2s_total + s2s.numel()
        self.t2t_total = self.t2t_total + t2t.numel()
        self.s2t_total = self.s2t_total + s2t.numel()

    def compute(self) -> torch.Tensor:
        return (
            self.s2s / self.s2s_total
            + self.t2t / self.t2t_total
            - 2 * (self.s2t / self.s2t_total)
        )


class JointMaximumMeanDiscrepancyLoss(torchmetrics.Metric):
    """The Joint Maximum Mean Discrepancy Loss (JMMD) is a distance measure between
    multiple pairs of arbitrary distributions.

    It is related to the MMD insofar as the distance of each distribution pair in a
    reproducing Hilbert kernel space (RHKS) is calculated and then multiplied before
    the discrepancy is computed.

    ```python
    joint_rhks(xs, ys) = prod(rhks(x, y) for x, y in zip(xs, xs))
    ```

    For more information see
    [MaximumMeanDiscrepancyLoss] [rul_adapt.loss.adaption.MaximumMeanDiscrepancyLoss].
    """

    is_differentiable = True
    higher_is_better = False
    full_state_update = False

    s2s: torch.Tensor
    t2t: torch.Tensor
    s2t: torch.Tensor

    s2s_total: torch.Tensor
    t2t_total: torch.Tensor
    s2t_total: torch.Tensor

    def __init__(self) -> None:
        """
        Create a new JMMD loss module.

        It features a single Gaussian kernel with a bandwidth chosen by the median
        heuristic.
        """
        super().__init__()

        self.add_state("s2s", torch.tensor(0.0), dist_reduce_fx="sum")
        self.add_state("t2t", torch.tensor(0.0), dist_reduce_fx="sum")
        self.add_state("s2t", torch.tensor(0.0), dist_reduce_fx="sum")

        self.add_state("s2s_total", torch.tensor(0.0), dist_reduce_fx="sum")
        self.add_state("t2t_total", torch.tensor(0.0), dist_reduce_fx="sum")
        self.add_state("s2t_total", torch.tensor(0.0), dist_reduce_fx="sum")

    def update(
        self, source_features: List[torch.Tensor], target_features: List[torch.Tensor]
    ) -> None:
        """
        Compute the JMMD loss between multiple feature distributions.

        Args:
            source_features: The list of source features of shape
                             `[batch_size, num_features]`.
            target_features: The list of target features of shape
                             `[batch_size, num_features]`.

        Returns:
            scalar JMMD loss
        """
        distances = []
        for source, target in zip(source_features, target_features):
            feats = torch.cat([source, target], dim=0)
            dist = calc_pairwise_euclidean(feats, feats)
            (gamma,) = _get_gammas(dist, 1)
            distances.append(_calc_gaussian_kernel(dist, gamma))

        batch_size = source_features[0].shape[0]
        merged_distances = torch.stack(distances, dim=0).prod(dim=0)
        s2s, t2t, s2t = _split_distances(merged_distances, batch_size)

        self.s2s = self.s2s + s2s.sum()
        self.t2t = self.t2t + t2t.sum()
        self.s2t = self.s2t + s2t.sum()

        self.s2s_total = self.s2s_total + s2s.numel()
        self.t2t_total = self.t2t_total + t2t.numel()
        self.s2t_total = self.s2t_total + s2t.numel()

    def compute(self) -> torch.Tensor:
        return (
            self.s2s / self.s2s_total
            + self.t2t / self.t2t_total
            - 2 * (self.s2t / self.s2t_total)
        )


class DomainAdversarialLoss(torchmetrics.Metric):
    """The Domain Adversarial Neural Network Loss (DANN) uses a domain discriminator
    to measure the distance between two feature distributions.

    The domain discriminator is a neural network that is jointly trained on
    classifying its input as one of two domains. Its output should be a single
    unscaled score (logit) which is fed to a binary cross entropy loss.

    The domain discriminator is preceded by a [GradientReversalLayer]
    [rul_adapt.loss.adaption.GradientReversalLayer]. This way, the discriminator is
    trained to separate the domains while the network generating the inputs is
    trained to marginalize the domain difference."""

    is_differentiable = True
    higher_is_better = False
    full_state_update = False

    loss: torch.Tensor
    total: torch.Tensor

    def __init__(self, domain_disc: nn.Module) -> None:
        """
        Create a new DANN loss module.

        Args:
            domain_disc: The neural network to act as the domain discriminator.
        """
        super().__init__()

        self.domain_disc = domain_disc
        self.grl = GradientReversalLayer()

        self.add_state("loss", torch.tensor(0.0), dist_reduce_fx="sum")
        self.add_state("total", torch.tensor(0.0), dist_reduce_fx="sum")

    def update(self, source: torch.Tensor, target: torch.Tensor) -> None:
        """
        Calculate the DANN loss as the binary cross entropy of the discriminators
        prediction for the source and target features.

        The source features receive a domain label of zero and the target features a
        domain label of one.

        Args:
            source: The source features with domain label zero.
            target: The target features with domain label one.
        """
        inputs = torch.cat([source, target])
        combined_batch_size = inputs.shape[0]

        labels = torch.ones(combined_batch_size, 1, device=self.device)
        source_batch_size = source.shape[0]
        labels[:source_batch_size] *= 0.0

        predictions = self.domain_disc(self.grl(inputs))
        loss = nn.functional.binary_cross_entropy_with_logits(
            predictions, labels, reduction="sum"
        )

        self.loss = self.loss + loss
        self.total = self.total + combined_batch_size

    def compute(self) -> torch.Tensor:
        return self.loss / self.total


class GradientReversalLayer(nn.Module):
    """The gradient reversal layer (GRL) acts as an identity function in the forward
    pass and scales the gradient by a negative scalar in the backward pass.

    ```python
    GRL(f(x)) = f(x)
    GRL`(f(x)) = -lambda * f`(x)
    ```
    """

    grad_weight: torch.Tensor

    def __init__(self, grad_weight: float = 1.0) -> None:
        """
        Create a new Gradient Reversal Layer.

        Args:
            grad_weight: The scalar that weights the negative gradient.
        """
        super().__init__()

        self.register_buffer("grad_weight", torch.tensor(grad_weight))

    def forward(self, inputs: torch.Tensor) -> torch.Tensor:
        return _gradient_reversal(inputs, self.grad_weight)


def _gradient_reversal(x: torch.Tensor, grad_weight: torch.Tensor) -> torch.Tensor:
    """Perform gradient reversal on input."""
    return _GradientReverse.apply(x, grad_weight)


class _GradientReverse(torch.autograd.Function):
    """Gradient reversal forward and backward definitions."""

    @staticmethod
    def forward(  # type: ignore
        ctx: Any, inputs: torch.Tensor, grad_weight: torch.Tensor
    ) -> torch.Tensor:
        """Forward pass as identity mapping."""
        ctx.grad_weight = grad_weight

        return inputs

    @staticmethod
    def backward(  # type: ignore
        ctx: Any, grad: torch.Tensor
    ) -> Tuple[torch.Tensor, None]:
        """Backward pass as negative, scaled gradient."""
        return -ctx.grad_weight * grad, None

    @staticmethod
    def jvp(ctx: Any, *grad_inputs: Any) -> Any:
        pass


def _split_distances(distances, split_idx):
    s2s = distances[:split_idx, :split_idx]
    t2t = distances[split_idx:, split_idx:]
    s2t = distances[:split_idx, split_idx:]

    return s2s, t2t, s2t


def _calc_multi_kernel(distances: torch.Tensor, gammas: List[float]) -> torch.Tensor:
    """Compute the linear combination of Gaussian kernels."""
    kernels = [_calc_gaussian_kernel(distances, gamma) for gamma in gammas]
    combined: torch.Tensor = sum(kernels) / len(gammas)  # type: ignore

    return combined


def _calc_gaussian_kernel(distances: torch.Tensor, gamma: float) -> torch.Tensor:
    """Compute a single Gaussian kernel."""
    return torch.exp(-distances * gamma)


def _get_gammas(distances: torch.Tensor, num_kernels: int) -> List[float]:
    """Compute gammas for n Gaussian kernels via median heuristic."""
    n_samples = distances.shape[0]
    bandwidth = (n_samples**2 - n_samples) / torch.sum(distances.detach())
    bandwidth /= 2 ** (num_kernels // 2)
    gammas = [bandwidth * (2**i) for i in range(num_kernels)]

    return gammas


class ConsistencyLoss(torchmetrics.Metric):
    is_differentiable = True
    higher_is_better = False
    full_state_update = False

    loss: List[torch.Tensor]
    total: List[torch.Tensor]

    def __init__(self):
        super().__init__()

        self.add_state("loss", [], dist_reduce_fx="cat")
        self.add_state("total", [], dist_reduce_fx="cat")

    def update(
        self, leader_features: torch.Tensor, follower_features: torch.Tensor
    ) -> None:
        loss = torch.mean(torch.abs(leader_features - follower_features))
        batch_size = torch.tensor(leader_features.shape[0], device=self.device)

        self.loss.append(loss)
        self.total.append(batch_size)

    def compute(self) -> torch.Tensor:
        return weighted_mean(self.loss, self.total)



import warnings
from itertools import tee
from typing import Union, Callable, Literal, Any, Optional, Iterable

import pytorch_lightning
import torch
import torchmetrics
from torch import nn

if pytorch_lightning.__version__.startswith("2."):
    from pytorch_lightning.utilities.types import OptimizerLRSchedulerConfig  # type: ignore
else:
    OptimizerLRSchedulerConfig = dict  # type: ignore


def pairwise(iterable):
    """s -> (s0,s1), (s1,s2), (s2, s3), ..."""
    a, b = tee(iterable)
    next(b, None)
    return zip(a, b)


def str2callable(cls: Union[str, Callable], restriction: str = "") -> Callable:
    """Dynamically import a callable from a string."""
    if isinstance(cls, Callable):  # type: ignore[arg-type]
        return cls  # type: ignore[return-value]
    if isinstance(cls, str) and not cls.startswith(restriction):
        raise ValueError(
            f"Failed to import '{cls}' because "
            f"imports are restricted to '{restriction}'."
        )

    module_name, class_name = cls.rsplit(".", 1)  # type: ignore[union-attr]
    module = __import__(module_name, fromlist=[class_name])
    cls = getattr(module, class_name)

    return cls  # type: ignore[return-value]


def get_loss(loss_type: str) -> torchmetrics.Metric:
    """Get a loss instance by specifying a string."""
    loss: torchmetrics.Metric
    if loss_type == "mae":
        loss = torchmetrics.MeanAbsoluteError()
    elif loss_type == "mse":
        loss = torchmetrics.MeanSquaredError()
    elif loss_type == "rmse":
        loss = torchmetrics.MeanSquaredError(squared=False)
    else:
        raise ValueError(
            f"Unknown loss type '{loss_type}'. " "Use either 'mae', 'mse' or 'rmse'."
        )

    return loss


def dataloader2domain(dataloader_idx: int) -> Literal["source", "target"]:
    if dataloader_idx == 0:
        return "source"
    elif dataloader_idx == 1:
        return "target"
    else:
        raise RuntimeError(
            f"Expected dataloader_idx to be 0 or 1 but got {dataloader_idx}."
        )


class OptimizerFactory:
    """Factory for creating optimizers and schedulers.

    After initialization, the factory can be called to create an optimizer with an
    optional scheduler."""

    def __init__(
        self,
        optim_type: str = "adam",
        lr: float = 1e-3,
        scheduler_type: Optional[str] = None,
        **kwargs: Any,
    ) -> None:
        """
        Create a new factory to efficiently create optimizers and schedulers.

        The factory creates an optimizer of the specified `optim_type` and adds an
        optional scheduler of the specified `scheduler_type`. Additional keyword
        arguments for the optimizer can be passed by adding the 'optim_' prefix and
        for the scheduler by adding the 'scheduler_' prefix. The factory will ignore
        any other keyword arguments.

        Available optimizers are 'adam', 'sgd' and 'rmsprop'. Available schedulers
        are 'step', 'cosine', 'linear' and 'lambda'.

        Args:
            optim_type: The type of optimizer to create.
            lr: The learning rate to use.
            scheduler_type: The optional type of scheduler to create.
            **kwargs: Additional keyword arguments for the optimizer and scheduler.
        """
        self.optim_type = optim_type
        self.lr = lr
        self.scheduler_type = scheduler_type
        self._kwargs = kwargs

        self._warn_excess_kwargs()

    def _warn_excess_kwargs(self) -> None:
        def _is_excess_kwarg(key: str) -> bool:
            return not (key.startswith("optim_") or key.startswith("scheduler_"))

        excess_kwargs = list(filter(_is_excess_kwarg, self._kwargs.keys()))
        if excess_kwargs:
            warnings.warn(
                "The following kwargs were passed but do not start "
                "with 'optim_' or 'scheduler_' and therefore "
                f"will be ignored: {excess_kwargs}."
            )

    def __call__(
        self, parameters: Iterable[nn.Parameter]
    ) -> OptimizerLRSchedulerConfig:
        """
        Create an optimizer with an optional scheduler for the given parameters.

        The object returned by this method is a lightning optimizer config and can be
        the return value of `configure_optimizers`.

        Args:
            parameters: The model parameters to optimize.

        Returns:
            A lightning optimizer config.
        """
        optim_kwargs = {
            key.replace("optim_", ""): value
            for key, value in self._kwargs.items()
            if key.startswith("optim_")
        }
        optim = self._optim_func(parameters, lr=self.lr, **optim_kwargs)
        optim_config = OptimizerLRSchedulerConfig(optimizer=optim)

        if self.scheduler_type is not None:
            scheduler_kwargs = {
                key.replace("scheduler_", ""): value
                for key, value in self._kwargs.items()
                if key.startswith("scheduler_")
            }
            optim_config["lr_scheduler"] = {
                "scheduler": self._scheduler_func(optim, **scheduler_kwargs)
            }

        return optim_config

    @property
    def _optim_func(self) -> Callable:
        if self.optim_type == "adam":
            return torch.optim.Adam
        elif self.optim_type == "sgd":
            return torch.optim.SGD
        elif self.optim_type == "rmsprop":
            return torch.optim.RMSprop
        else:
            raise ValueError(
                f"Unknown optimizer type '{self.optim_type}'. "
                "Use either 'adam', 'sgd' or 'rmsprop'."
            )

    @property
    def _scheduler_func(self) -> Callable:
        if self.scheduler_type == "step":
            return torch.optim.lr_scheduler.StepLR
        elif self.scheduler_type == "cosine":
            return torch.optim.lr_scheduler.CosineAnnealingLR
        elif self.scheduler_type == "linear":
            return torch.optim.lr_scheduler.LinearLR
        elif self.scheduler_type == "lambda":
            return torch.optim.lr_scheduler.LambdaLR
        else:
            raise ValueError(
                f"Unknown scheduler type '{self.scheduler_type}'. "
                "Use either 'step', 'cosine', 'linear' or 'lambda'."
            )



"""The TBiGRU approach uses a feature selection mechanism to mine transferable
features and a bearing running state detection to determine the
first-time-to-predict. The training is done with an [MMD approach]
[rul_adapt.approach.mmd].

The feature selection uses a distance measure based on Dynamic Time Warping and the
Wasserstein distance. From a set of 30 common vibration features the ones with the
smallest distance between source and target domain are selected. These features serve
as inputs to the network.

The first-time-to-predict (FTTP) is used to generate the RUL labels for training.
FTTP is the time step where the degradation can be detected for the first time. The
RUL labels before this time step should be constant. The TBiGRU approach uses the
moving average correlation (MAC) of the energy entropies of four levels of maximal
overlap discrete wavelet transform (MODWT) decompositions to determine four running
states of each bearing. The end of the steady running state marks the FTTP.

TBiGRU was introduced by [Cao et al.](
https://doi.org/10.1016/j.measurement.2021.109287) and evaluated on the FEMTO Bearing
dataset."""

from itertools import product
from queue import Queue
from typing import List, Tuple, Optional

import numpy as np
import pywt  # type: ignore
import scipy.stats  # type: ignore
from dtaidistance import dtw  # type: ignore
from rul_datasets.reader import AbstractReader
from rul_datasets.utils import extract_windows
from scipy.stats import wasserstein_distance  # type: ignore
from sklearn.preprocessing import MinMaxScaler  # type: ignore


def rms(inputs: np.ndarray) -> np.ndarray:
    return np.sqrt(np.mean(inputs**2, axis=-2))


def p2p(inputs: np.ndarray) -> np.ndarray:
    return np.max(inputs, axis=-2) - np.min(inputs, axis=-2)


def skewness(inputs: np.ndarray) -> np.ndarray:
    sdev = inputs.std(axis=-2)
    mean = inputs.mean(axis=-2, keepdims=True)
    skew = np.sum((inputs - mean) ** 3, axis=-2) / (inputs.shape[-2] * sdev**3)

    return skew


def impulse_factor(inputs: np.ndarray) -> np.ndarray:
    absolute = np.abs(inputs)
    imp_f = np.max(absolute, axis=-2) / np.mean(absolute, axis=-2)

    return imp_f


def median_absolute(inputs: np.ndarray) -> np.ndarray:
    return np.median(np.abs(inputs), axis=-2)


def mean_absolute(inputs: np.ndarray) -> np.ndarray:
    return np.mean(np.abs(inputs), axis=-2)


def max_absolute(inputs: np.ndarray) -> np.ndarray:
    return np.max(np.abs(inputs), axis=-2)


def kurtosis(inputs: np.ndarray) -> np.ndarray:
    return scipy.stats.kurtosis(inputs, axis=-2, fisher=False)


def std(inputs: np.ndarray) -> np.ndarray:
    return np.std(inputs, axis=-2)


def margin_factor(inputs: np.ndarray) -> np.ndarray:
    absolute = np.abs(inputs)
    mf = np.max(absolute, axis=-2) / (np.mean(np.sqrt(absolute), axis=-2)) ** 2

    return mf


def energy(inputs: np.ndarray) -> np.ndarray:
    return np.sum(inputs**2, axis=-2)


def gini_factor(inputs: np.ndarray) -> np.ndarray:
    batched = len(inputs.shape) == 3
    inputs = inputs if batched else inputs[None]
    gini = _approx_batched_gini_factor(inputs)
    gini = gini if batched else gini.squeeze(0)

    return gini


def _approx_batched_gini_factor(inputs: np.ndarray) -> np.ndarray:
    window_size = inputs.shape[1]
    inputs = np.sort(inputs, axis=1)
    cumsum = np.cumsum(inputs, axis=1)
    gini = (window_size + 1 - 2 * np.sum(cumsum, axis=1) / cumsum[:, -1]) / window_size

    return gini


def band_energies(inputs: np.ndarray) -> np.ndarray:
    wp = pywt.WaveletPacket(inputs, wavelet="dmey", maxlevel=4, axis=-2)
    bands = (node.data for node in wp.get_level(4))
    energies = np.concatenate([energy(b) for b in bands], axis=-1)

    return energies


def std_ihc(inputs: np.ndarray) -> np.ndarray:
    return np.std((np.arccosh(inputs + 0j)), axis=-2)


def std_ihs(inputs: np.ndarray) -> np.ndarray:
    return np.std(np.arcsinh(inputs), axis=-2)


class VibrationFeatureExtractor:
    """This class extracts 30 different features from a raw acceleration signal.

    The features are: RMS, kurtosis, peak2peak, standard deviation, skewness,
    margin factor, impulse factor, energy, median absolute, gini factor, maximum
    absolute, mean absolute, energies of the 16 bands resulting from wavelet packet
    decomposition, standard deviation of arccosh and arcsinh. If the input has n
    features, n*30 features are extracted. Additionally, it features a scaler that
    can be fit to scale all extracted features between [0, 1]."""

    _scaler: Optional[MinMaxScaler]

    def __init__(
        self, num_input_features: int, feature_idx: Optional[List[int]] = None
    ) -> None:
        """
        Create a new vibration feature extractor with the selected features.

        The features are sorted as f1_1, .., f1_j, ..., fi_j, where i is the index of
        the computed feature (between 0 and 30) and j is the index of the raw
        feature (between 0 and `num_input_features`).

        Args:
            num_input_features: The number of input features.
            feature_idx: The indices of the features to compute.
        """
        self.num_input_features = num_input_features
        self.feature_idx = list(range(60)) if feature_idx is None else feature_idx

        if min(self.feature_idx) < 0 or max(self.feature_idx) > 60:
            raise ValueError("Feature indices need to be between 0 and 60.")

        self._scaler = None

    def __call__(
        self, features: np.ndarray, targets: np.ndarray
    ) -> Tuple[np.ndarray, np.ndarray]:
        """
        Extract the features from the input and optionally scale them.

        The features should have the shape `[num_windows, window_size,
        num_input_features]` and the targets `[num_windows]`.

        Args:
            features: The input features.
            targets: The input targets.

        Returns:
            The extracted features and input targets.
        """
        features = self._extract_selected(features)
        if self._scaler is not None:
            features = self._scaler.transform(features)

        return features, targets

    def _extract_selected(self, features: np.ndarray) -> np.ndarray:
        return _extract_all(features, self.num_input_features)[:, self.feature_idx]

    def fit(self, features: List[np.ndarray]) -> "VibrationFeatureExtractor":
        """
        Fit the internal scaler on a list of raw feature time series.

        The time series are passed through the feature extractor and then used to fit
        the internal min-max scaler. Each time series in the list should have the
        shape `[num_windows, window_size, num_input_features]`.

        Args:
            features: The list of raw feature time series.

        Returns:
            The feature extractor itself.
        """
        features = [self._extract_selected(f) for f in features]
        self._scaler = MinMaxScaler()
        for feat in features:
            self._scaler.partial_fit(feat)

        return self


def _extract_all(features: np.ndarray, num_features: int) -> np.ndarray:
    feature_list = [
        rms(features),
        kurtosis(features),
        p2p(features),
        std(features),
        skewness(features),
        margin_factor(features),
        impulse_factor(features),
        energy(features),
        median_absolute(features),
        gini_factor(features),
        max_absolute(features),
        mean_absolute(features),
        *band_energies(features)
        .reshape((-1, 16, num_features))
        .transpose(1, 0, 2),  # unpack features
        std_ihc(features),
        std_ihs(features),
    ]
    features = np.concatenate(feature_list, axis=1)

    return features


def select_features(
    source: AbstractReader, target: AbstractReader, num_features: int
) -> List[int]:
    """
    Select the most transferable features between source and target domain.

    30 features are considered: RMS, kurtosis, peak2peak, standard deviation, skewness,
    margin factor, impulse factor, energy, median absolute, gini factor, maximum
    absolute, mean absolute, energies of the 16 bands resulting from wavelet packet
    decomposition, standard deviation of arccosh and arcsinh. If the input has n raw
    features, n*30 features are extracted.

    The `dev` splits of both domains are used to calculate a distance metric based on
    Dynamic Time Warping and the Wasserstein Distance. The indices of the
    `num_feature` features with the lowest distances are returned.

    Args:
        source: The reader of the source domain.
        target: The reader of the target domain.
        num_features: The number of transferable features to return.

    Returns:
        The indices of features ordered by transferability.
    """
    source.prepare_data()
    target.prepare_data()
    source_runs, _ = source.load_split("dev")
    target_runs, _ = target.load_split("dev")

    num_org_features = source_runs[0].shape[-1]
    distances = np.zeros(30 * num_org_features)
    for source_run, target_run in product(source_runs, target_runs):
        source_feats = _extract_all(source_run, num_features=num_org_features)
        target_feats = _extract_all(target_run, num_features=num_org_features)
        for i, (s, f) in enumerate(zip(source_feats.T, target_feats.T)):
            distances[i] += _domain_distance(s, f)

    feature_idx = np.argsort(distances)[:num_features].tolist()

    return feature_idx


def _domain_distance(source: np.ndarray, target: np.ndarray, ratio=0.4) -> float:
    dtw_dist = dtw.distance_fast(source, target)
    wasserstein_dist = wasserstein_distance(source, target)
    distance = ratio * dtw_dist + (1 - ratio) * wasserstein_dist

    return distance


def mac(inputs: np.ndarray, window_size: int, wavelet: str = "dmey") -> np.ndarray:
    """
    Calculate the moving average correlation (MAC) of the energy entropies of four
    levels of maximal overlap discrete wavelet transform (MODWT) decompositions.

    The `wavelet` is a wavelet description that can be passed to `pywt`. The default
    wavelet was confirmed by the original authors. For more options call
    `pywt.wavelist`. The input signal should have the shape `[num_windows,
    window_size, num_features]`.

    Args:
        inputs: The input acceleration signal.
        window_size: The window size of the sliding window to calculate the average
                     over.
        wavelet: The description of the wavelet, e.g. 'sym4'.

    Returns:
        The MAC of the input signal which is `window_size - 1` shorter.
    """
    entropies = _energy_entropies(inputs, wavelet)
    entropies = extract_windows(entropies, window_size)
    anchor, queries = entropies[:, -2:-1], entropies[:, :-1]
    corr = _pearson(anchor, queries)
    corr = np.mean(np.abs(corr), axis=1)

    return corr


def _energy_entropies(inputs: np.ndarray, wavelet: str = "sym4") -> np.ndarray:
    coeffs = modwpt(inputs, wavelet, 4)
    energies = energy(coeffs)
    ratios = energies / np.sum(energies, axis=-1, keepdims=True)
    entropy = -np.sum(ratios * np.log(ratios), axis=-1, keepdims=True)
    entropies = ratios / entropy

    return entropies


def _pearson(x: np.ndarray, y: np.ndarray) -> np.ndarray:
    diff_x = x - np.mean(x, axis=2, keepdims=True)
    diff_y = y - np.mean(y, axis=2, keepdims=True)
    cov = np.sum(diff_y * diff_x, axis=2)
    std_product = np.sqrt(np.sum(diff_y**2, axis=2) * np.sum(diff_x**2, axis=2))
    corr = cov / std_product

    return corr


def modwpt(inputs: np.ndarray, wavelet: str, level: int) -> np.ndarray:
    """
    Apply Maximal Overlap Discrete Wavelet Packet Transformation (MODWT) of `level`
    to the input.

    The `wavelet` should be a string that can be passed to `pywt` to construct a
    wavelet function. For more options call `pywt.wavelist`. The implementation was
    inspired by [this repository](https://github.com/pistonly/modwtpy).

    Args:
        inputs: An input signal of shape `[num_windows, window_size, num_features]`.
        wavelet: The description of the wavelet function, e.g. 'sym4'.
        level: The decomposition level.

    Returns:
        The 2**level decompositions stacked in the last axis.
    """
    if level < 1:
        raise ValueError("The level needs to be a positive integer.")
    wavelet_func = pywt.Wavelet(wavelet)
    dec_hi = np.array(wavelet_func.dec_hi) / np.sqrt(2)
    dec_lo = np.array(wavelet_func.dec_lo) / np.sqrt(2)

    input_queue: Queue = Queue(maxsize=2**level)
    input_queue.put(inputs)
    for j in range(level):
        coeffs = _decompose_level(input_queue, dec_hi, dec_lo, j)
        for d in coeffs:
            input_queue.put(d)

    return np.concatenate(coeffs, axis=-1)


def _decompose_level(input_queue, dec_hi, dec_lo, level):
    coeffs = []
    while not input_queue.empty():
        signal = input_queue.get()
        detail = _circular_convolve_fast(dec_hi, signal, level + 1)
        approx = _circular_convolve_fast(dec_lo, signal, level + 1)
        coeffs.append(approx)
        coeffs.append(detail)

    return coeffs


def _circular_convolve_d(
    kernel: np.ndarray, signal: np.ndarray, level: int
) -> np.ndarray:
    len_signal = len(signal)
    len_wavelet = len(kernel)
    convolved = np.zeros(len_signal)
    wavelet_range = np.arange(len_wavelet)
    for t in range(len_signal):
        index = np.mod(t - 2 ** (level - 1) * wavelet_range, len_signal)
        element = np.array([signal[ind] for ind in index])
        convolved[t] = (np.array(kernel) * element).sum()

    return convolved


def _circular_convolve_fast(
    kernel: np.ndarray, signal: np.ndarray, level: int
) -> np.ndarray:
    len_signal = signal.shape[-2]
    signal_range = np.arange(len_signal)[:, None]
    wavelet_range = np.arange(len(kernel))[None, :]
    idx = np.mod(signal_range - 2 ** (level - 1) * wavelet_range, len_signal)
    convolved = np.sum(kernel[None, :, None] * signal[..., idx, :], axis=-2)

    return convolved



from typing import Callable, List

import torch


def weighted_mean(
    inputs: List[torch.Tensor], weights: List[torch.Tensor]
) -> torch.Tensor:
    torch_weights = torch.stack(weights).float()
    torch_weights /= torch.sum(torch_weights)
    mean = torch.sum(torch.stack(inputs) * torch_weights)

    return mean


def calc_pairwise_euclidean(x: torch.Tensor, y: torch.Tensor) -> torch.Tensor:
    return _calc_pairwise_distances(x, y, _euclidean)


def calc_pairwise_dot(x: torch.Tensor, y: torch.Tensor) -> torch.Tensor:
    return _calc_pairwise_distances(x, y, _dot)


def _calc_pairwise_distances(
    x: torch.Tensor, y: torch.Tensor, dist_func: Callable
) -> torch.Tensor:
    """Compute pairwise euclidean distances between features."""
    x = x[:, None]
    y = y[None, :]
    distances = dist_func(x, y)

    return distances


def _euclidean(x: torch.Tensor, y: torch.Tensor) -> torch.Tensor:
    return torch.sqrt((x - y) ** 2 + 1e-7).sum(-1)


def _dot(x, y):
    return torch.mul(x, y).sum(-1)



from typing import Literal, Tuple, Any, Optional

import hydra
import omegaconf
import pytorch_lightning as pl
import rul_datasets

from rul_adapt.approach import LatentAlignApproach


def get_latent_align(
    dataset: Literal["cmapss", "xjtu-sy"],
    source_fd: int,
    target_fd: int,
    xjtu_sy_subtask: Optional[int] = None,
    **trainer_kwargs: Any,
) -> Tuple[rul_datasets.LatentAlignDataModule, LatentAlignApproach, pl.Trainer]:
    """
    Construct a Latent Alignment approach for the selected dataset with the original
    hyperparameters.

    For the XJTU-SY task only FD001 and FD002 are available. The subtask controls if
    the bearing with the id 1 or 2 is used as the target data.

    Examples:
        ```pycon
        >>> import rul_adapt
        >>> dm, latent, trainer = rul_adapt.construct.get_latent_align("cmapss", 3, 1)
        >>> trainer.fit(latent, dm)
        >>> trainer.test(latent, dm)
        ```

    Args:
        dataset: The dataset to use.
        source_fd: The source FD.
        target_fd: The target FD.
        xjtu_sy_subtask: The subtask for the XJTU-SY (either 1 or 2).
        trainer_kwargs: Overrides for the trainer class.
    Returns:
        dm: The data module for adaption of the sub-datasets.
        dann: The Latent Alignment approach with feature extractor and regressor.
        trainer: The trainer object.
    """
    config = get_latent_align_config(dataset, source_fd, target_fd, xjtu_sy_subtask)
    dm, latent_align, trainer = latent_align_from_config(config, **trainer_kwargs)

    return dm, latent_align, trainer


def get_latent_align_config(
    dataset: Literal["cmapss", "xjtu-sy"],
    source_fd: int,
    target_fd: int,
    xjtu_sy_subtask: Optional[int] = None,
) -> omegaconf.DictConfig:
    """
    Get a configuration for the Latent Alignment approach.

    For the XJTU-SY task only FD001 and FD002 are available. The subtask controls if
    the bearing with the id 1 or 2 is used as the target data. The configuration can
    be modified and fed to [latent_align_from_config]
    [rul_adapt.construct.latent_align.latent_align_from_config] to create the approach.

    Args:
        dataset: The dataset to use.
        source_fd: The source FD.
        target_fd: The target FD.
        xjtu_sy_subtask: The subtask for the XJTU-SY (either 1 or 2).
    Returns:
        The Latent Alignment configuration.
    """
    _validate(dataset, source_fd, target_fd, xjtu_sy_subtask)
    overrides = [
        f"+dataset={dataset}",
        f"dm.source.fd={source_fd}",
        f"dm.target.fd={target_fd}",
    ]
    if dataset == "xjtu-sy":
        overrides.append(f"+subtask=SUB{target_fd}{xjtu_sy_subtask}")
    else:
        overrides.append(f"+split_steps=FD00{target_fd}")
    with hydra.initialize("config", version_base="1.1"):
        config = hydra.compose("base", overrides)

    return config


def latent_align_from_config(
    config: omegaconf.DictConfig, **trainer_kwargs: Any
) -> Tuple[rul_datasets.LatentAlignDataModule, LatentAlignApproach, pl.Trainer]:
    """
    Construct a Latent Alignment approach from a configuration.

    The configuration can be created by calling [get_latent_align_config]
    [rul_adapt.construct.latent_align.get_latent_align_config].

    Args:
        config: The Latent Alignment configuration.
        trainer_kwargs: Overrides for the trainer class.
    Returns:
        dm: The data module for adaption of the sub-datasets.
        dann: The Latent Alignment approach with feature extractor, regressor.
        trainer: The trainer object.
    """
    source = hydra.utils.instantiate(config.dm.source)
    target = hydra.utils.instantiate(config.dm.target)
    kwargs = hydra.utils.instantiate(config.dm.kwargs)
    dm = rul_datasets.LatentAlignDataModule(
        rul_datasets.RulDataModule(source, **kwargs),
        rul_datasets.RulDataModule(target, **kwargs),
        **config.dm.adaption_kwargs,
    )

    feature_extractor = hydra.utils.instantiate(config.feature_extractor)
    regressor = hydra.utils.instantiate(config.regressor)

    approach = hydra.utils.instantiate(config.latent_align)
    approach.set_model(feature_extractor, regressor)

    trainer = hydra.utils.instantiate(config.trainer, **trainer_kwargs)

    return dm, approach, trainer


def _validate(
    dataset: Literal["cmapss", "xjtu-sy"],
    source_fd: int,
    target_fd: int,
    xjtu_sy_subtask: Optional[int],
) -> None:
    if dataset not in ["cmapss", "xjtu-sy"]:
        raise ValueError(f"No configuration for '{dataset}'.")
    elif source_fd == target_fd:
        raise ValueError(
            f"No configuration for adapting from FD{source_fd:03} to itself."
        )
    elif dataset == "cmapss":
        _validate_cmapss(source_fd, target_fd)
    elif dataset == "xjtu-sy":
        _validate_xjtu_sy(source_fd, target_fd, xjtu_sy_subtask)


def _validate_cmapss(source_fd: int, target_fd: int):
    if 1 > source_fd or source_fd > 4:
        raise ValueError(f"CMAPSS has only FD001 to FD004 but no FD{source_fd:03}")
    elif 1 > target_fd or target_fd > 4:
        raise ValueError(f"CMAPSS has only FD001 to FD004 but no FD{target_fd:03}")


def _validate_xjtu_sy(source_fd: int, target_fd: int, subtask: Optional[int]):
    if 1 > source_fd or source_fd > 2:
        raise ValueError(
            "Only FD001 and FD002 of XJTU-SY are used in "
            f"this approach but not FD{source_fd:03}."
        )
    elif 1 > target_fd or target_fd > 2:
        raise ValueError(
            "Only FD001 and FD002 of XJTU-SY are used in "
            f"this approach but not FD{target_fd:03}."
        )
    elif subtask is None:
        raise ValueError("XJTU-SY requires a subtask of 1 or 2.")
    elif 1 > subtask or subtask > 2:
        raise ValueError(
            f"XJTU-SY has only subtasks 1 and 2 but not subtask {subtask}."
        )



"""These losses are used to create a latent space that is conductive to RUL
estimation. They are mainly used by the [LatentAlignmentApproach]
[rul_adapt.approach.LatentAlignApproach]."""

from typing import List

import torch
import torchmetrics

from rul_adapt.loss.utils import calc_pairwise_dot, weighted_mean


class HealthyStateAlignmentLoss(torchmetrics.Metric):
    """
    This loss is used to align the healthy state of the data in the latent space.

    It computes the mean of the variance of each latent feature which is supposed to
    be minimized. This way a single compact cluster of healthy state data should be
    formed.

    The loss is implemented as a [torchmetrics.Metric](
    https://torchmetrics.readthedocs.io/en/stable/pages/quickstart.html#module
    -metrics). See their documentation for more information.

    Examples:
        ```pycon
        >>> from rul_adapt.loss.alignment import HealthyStateAlignmentLoss
        >>> healthy_align = HealthyStateAlignmentLoss()
        >>> healthy_align(torch.zeros(10, 5))
        tensor(0.)
        ```
    """

    is_differentiable: bool = True
    higher_is_better: bool = False
    full_state_update: bool = False

    loss: List[torch.Tensor]
    total: List[torch.Tensor]

    def __init__(self):
        super().__init__()

        self.add_state("loss", [], dist_reduce_fx="cat")
        self.add_state("total", [], dist_reduce_fx="cat")

    def update(self, healthy: torch.Tensor) -> None:
        self.loss.append(torch.mean(torch.var(healthy, dim=0)))
        self.total.append(torch.tensor(healthy.shape[0], device=self.device))

    def compute(self) -> torch.Tensor:
        return weighted_mean(self.loss, self.total)


class DegradationDirectionAlignmentLoss(torchmetrics.Metric):
    """
    This loss is used to align the direction of the degradation data in relation to
    the healthy state data in the latent space.

    It computes the mean of the cosine of the pairwise-angle of the vectors from the
    healthy state cluster to each degradation data point. The healthy state cluster
    location is assumed to be the mean of the healthy state data in the latent space.
    The loss is negated in order to maximize the cosine by minimizing the loss.

    The loss is implemented as a [torchmetrics.Metric](
    https://torchmetrics.readthedocs.io/en/stable/pages/quickstart.html#module
    -metrics). See their documentation for more information.

    Examples:
        ```pycon
        >>> from rul_adapt.loss.alignment import DegradationDirectionAlignmentLoss
        >>> degradation_align = DegradationDirectionAlignmentLoss()
        >>> degradation_align(torch.zeros(10, 5), torch.ones(10, 5))
        tensor(-1.0000)
        ```
    """

    is_differentiable: bool = True
    higher_is_better: bool = False
    full_state_update: bool = False

    loss: torch.Tensor
    total: torch.Tensor

    def __init__(self):
        super().__init__()

        self.add_state("loss", torch.tensor(0.0), dist_reduce_fx="sum")
        self.add_state("total", torch.tensor(0.0), dist_reduce_fx="sum")

    def update(self, healthy: torch.Tensor, degraded: torch.Tensor) -> None:
        healthy_mean = healthy.mean(dim=0)
        trajectory = degraded - healthy_mean
        trajectory_norm = torch.norm(trajectory, dim=1, keepdim=True)
        trajectory = trajectory / (trajectory_norm + 1e-8)
        pairwise_dist = calc_pairwise_dot(trajectory, trajectory)
        loss = -pairwise_dist.sum()

        self.loss = self.loss + loss
        self.total = self.total + (degraded.shape[0] ** 2)

    def compute(self) -> torch.Tensor:
        return self.loss / self.total


class DegradationLevelRegularizationLoss(torchmetrics.Metric):
    """
    This loss is used to regularize the degradation level of the data in the latent
    space in relation to the healthy state data.

    It computes the mean of the difference between the normalized distance of the
    degradation data points from the healthy state cluster and the normalized
    degradation steps. The healthy state cluster location is assumed to be the mean
    of the healthy state data in the latent space.

    The loss is implemented as a [torchmetrics.Metric](
    https://torchmetrics.readthedocs.io/en/stable/pages/quickstart.html#module
    -metrics). See their documentation for more information.

    Examples:
        ```pycon
        >>> from rul_adapt.loss.alignment import DegradationLevelRegularizationLoss
        >>> degradation_align = DegradationLevelRegularizationLoss()
        >>> degradation_align(torch.zeros(10, 5),
        ...                   torch.ones(10, 5),
        ...                   torch.ones(10),
        ...                   torch.ones(10, 5),
        ...                   torch.ones(10))
        tensor(0.)
        ```
    """

    is_differentiable: bool = True
    higher_is_better: bool = False
    full_state_update: bool = False

    loss: torch.Tensor
    total: torch.Tensor

    def __init__(self) -> None:
        super().__init__()

        self.add_state("loss", torch.tensor(0.0), dist_reduce_fx="sum")
        self.add_state("total", torch.tensor(0.0), dist_reduce_fx="sum")

    def update(
        self,
        healthy: torch.Tensor,
        source: torch.Tensor,
        source_degradation_steps: torch.Tensor,
        target: torch.Tensor,
        target_degradation_steps: torch.Tensor,
    ) -> None:
        healthy_mean = healthy.mean(dim=0)
        source_distances = self._calc_normed_distances(healthy_mean, source)
        target_distances = self._calc_normed_distances(healthy_mean, target)

        source_max = torch.max(source_degradation_steps)
        source_degradation_steps = source_degradation_steps / source_max
        target_max = torch.max(target_degradation_steps)
        target_degradation_steps = target_degradation_steps / target_max

        source_loss = torch.abs(source_distances - source_degradation_steps).sum()
        target_loss = torch.abs(target_distances - target_degradation_steps).sum()
        loss = source_loss + target_loss

        self.loss = self.loss + loss
        self.total = self.total + source.shape[0]

    def compute(self) -> torch.Tensor:
        return self.loss / self.total

    def _calc_normed_distances(
        self, healthy_mean: torch.Tensor, source: torch.Tensor
    ) -> torch.Tensor:
        distances = torch.norm(source - healthy_mean, p=2, dim=1)
        distances = distances / (torch.max(distances) + 1e-8)

        return distances
